from __future__ import annotations

import pytest

from diracx.db.sql.auth.db import AuthDB
from diracx.db.sql.auth.schema import RefreshTokenStatus


@pytest.fixture
async def auth_db(tmp_path):
    auth_db = AuthDB("sqlite+aiosqlite:///:memory:")
    async with auth_db.engine_context():
        async with auth_db.engine.begin() as conn:
            await conn.run_sync(auth_db.metadata.create_all)
        yield auth_db


async def test_insert(auth_db: AuthDB):
    """Insert two refresh tokens in the DB and check that they don't share the same JWT ID"""

    # Insert a first refresh token
    async with auth_db as auth_db:
        jti1, _ = await auth_db.insert_refresh_token(
            "subject",
            "username",
            "vo:lhcb property:NormalUser",
        )

    # Insert a second refresh token
    async with auth_db as auth_db:
        jti2, _ = await auth_db.insert_refresh_token(
            "subject",
            "username",
            "vo:lhcb property:NormalUser",
        )

    # Make sure they don't have the same JWT ID
    assert jti1 != jti2


async def test_get(auth_db: AuthDB):
    """Insert a refresh token in the DB and get it"""
    # Refresh token details we want to insert
    refresh_token_details = {
        "sub": "12345",
        "preferred_username": "John Doe",
        "scope": "vo:lhcb property:NormalUser",
    }

    # Insert refresh token details
    async with auth_db as auth_db:
        jti, creation_time = await auth_db.insert_refresh_token(
            refresh_token_details["sub"],
            refresh_token_details["preferred_username"],
            refresh_token_details["scope"],
        )

    # Enrich the dict with the generated refresh token attributes
    refresh_token_details["jti"] = jti
    refresh_token_details["status"] = RefreshTokenStatus.CREATED
    refresh_token_details["creation_time"] = creation_time

    # Get refresh token details
    async with auth_db as auth_db:
        result = await auth_db.get_refresh_token(jti)

    # Make sure they are identical
    assert result == refresh_token_details


async def test_get_user_refresh_tokens(auth_db: AuthDB):
    """Insert refresh tokens belonging to different users in the DB and
    get the refresh tokens of each user
    """
    # Two users
    sub1 = "subject1"
    sub2 = "subject2"

    # Insert tokens
    # - 2 of them belongs to sub1
    # - 1 of them belongs to sub2
    subjects = [sub1, sub1, sub2]
    async with auth_db as auth_db:
        for sub in subjects:
            await auth_db.insert_refresh_token(
                sub,
                "username",
                "scope",
            )

    # Get the refresh tokens of each user
    async with auth_db as auth_db:
        refresh_tokens_user1 = await auth_db.get_user_refresh_tokens(sub1)
        refresh_tokens_user2 = await auth_db.get_user_refresh_tokens(sub2)

    # Check the number of refresh tokens belonging to the users
    # And check that the subject value corresponds to the user's subject
    assert len(refresh_tokens_user1) == 2
    for refresh_token in refresh_tokens_user1:
        assert refresh_token["sub"] == sub1

    assert len(refresh_tokens_user2) == 1
    for refresh_token in refresh_tokens_user2:
        assert refresh_token["sub"] == sub2


async def test_revoke(auth_db: AuthDB):
    """Insert a refresh token in the DB, revoke it, and make sure it appears as REVOKED in the db"""
    # Insert a refresh token details
    async with auth_db as auth_db:
        jti, _ = await auth_db.insert_refresh_token(
            "subject",
            "username",
            "scope",
        )

    # Revoke the token
    async with auth_db as auth_db:
        await auth_db.revoke_refresh_token(jti)

    # Make sure it is revoked
    async with auth_db as auth_db:
        refresh_token_details = await auth_db.get_refresh_token(jti)

    assert refresh_token_details["status"] == RefreshTokenStatus.REVOKED


async def test_revoke_user_refresh_tokens(auth_db: AuthDB):
    """Insert refresh tokens in the DB, revoke them, and make sure it appears as REVOKED in the db"""
    # Two users
    sub1 = "subject1"
    sub2 = "subject2"

    # Insert tokens
    # - 2 of them belongs to sub1
    # - 1 of them belongs to sub2
    subjects = [sub1, sub1, sub2]
    async with auth_db as auth_db:
        for sub in subjects:
            await auth_db.insert_refresh_token(
                sub,
                "username",
                "scope",
            )

    # Revoke the tokens of sub1
    async with auth_db as auth_db:
        await auth_db.revoke_user_refresh_tokens(sub1)

    # Make sure they are revoked (but not the ones belonging to sub2)
    async with auth_db as auth_db:
        refresh_token_details = await auth_db.get_user_refresh_tokens(sub1)
        assert len(refresh_token_details) == 0
        refresh_token_details = await auth_db.get_user_refresh_tokens(sub2)
        assert len(refresh_token_details) == 1

    # Revoke the tokens of sub2
    async with auth_db as auth_db:
        await auth_db.revoke_user_refresh_tokens(sub2)

    # Make sure they are all revoked
    async with auth_db as auth_db:
        refresh_token_details = await auth_db.get_user_refresh_tokens(sub1)
        assert len(refresh_token_details) == 0
        refresh_token_details = await auth_db.get_user_refresh_tokens(sub2)
        assert len(refresh_token_details) == 0


async def test_revoke_and_get_user_refresh_tokens(auth_db: AuthDB):
    """Insert refresh tokens belonging to a user, revoke one of them and
    make sure that only the active tokens appear
    """
    # User
    sub = "subject"

    # Number of tokens to insert
    nb_tokens = 2

    # Insert tokens
    jtis = []
    async with auth_db as auth_db:
        for _ in range(nb_tokens):
            jti, _ = await auth_db.insert_refresh_token(
                sub,
                "username",
                "scope",
            )
            jtis.append(jti)

    # Get the refresh tokens of the user
    async with auth_db as auth_db:
        refresh_tokens_user = await auth_db.get_user_refresh_tokens(sub)

    # Check the number of refresh tokens belonging to the user
    # And check that the subject value corresponds to the user's subject
    assert len(refresh_tokens_user) == nb_tokens
    for refresh_token in refresh_tokens_user:
        assert refresh_token["sub"] == sub

    # Revoke one of the tokens
    async with auth_db as auth_db:
        await auth_db.revoke_refresh_token(jtis[0])

    # Get the refresh tokens of the user again
    async with auth_db as auth_db:
        refresh_tokens_user = await auth_db.get_user_refresh_tokens(sub)

    # Check that there is less refresh tokens returned
    # And check that the subject value corresponds to the user's subject
    assert len(refresh_tokens_user) == nb_tokens - 1
    for refresh_token in refresh_tokens_user:
        assert refresh_token["sub"] == sub
        assert refresh_token["jti"] != jtis[0]


async def test_get_refresh_tokens(auth_db: AuthDB):
    """Insert refresh tokens belonging to different users in the DB and
    get the refresh tokens
    """
    # Two users
    sub1 = "subject1"
    sub2 = "subject2"

    # Insert tokens
    # - 2 of them belongs to sub1
    # - 1 of them belongs to sub2
    subjects = [sub1, sub1, sub2]
    async with auth_db as auth_db:
        for sub in subjects:
            await auth_db.insert_refresh_token(
                sub,
                "username",
                "scope",
            )

    # Get all refresh tokens (Admin)
    async with auth_db as auth_db:
        refresh_tokens = await auth_db.get_user_refresh_tokens()

    # Check the number of retrieved refresh tokens (should be 3 refresh tokens)
    assert len(refresh_tokens) == 3

    # Get user refresh tokens (sub1)
    async with auth_db as auth_db:
        refresh_tokens = await auth_db.get_user_refresh_tokens(sub1)

    # Check the number of retrieved refresh tokens (should be 3 refresh tokens)
    assert len(refresh_tokens) == 2



from __future__ import annotations

from datetime import datetime, timezone

import opensearchpy
import pytest

from diracx.testing.osdb import DummyOSDB

DUMMY_DOCUMENT = {
    "DateField": datetime.now(tz=timezone.utc),
    "IntField": 1234,
    "KeywordField1": "keyword1",
    "KeywordField2": "keyword two",
    "TextField": "text value",
}


async def test_applies_new_indices(dummy_opensearch_db: DummyOSDB):
    """Ensure that the index template is applied to new indices."""
    index_mappings = await _get_test_index_mappings(dummy_opensearch_db)
    # Ensure the index template was applied during index creation
    assert index_mappings == {"properties": dummy_opensearch_db.fields}


async def dummy_opensearch_db_without_template(dummy_opensearch_db: DummyOSDB):
    """Sanity test that previous test fails if there isn't a template."""
    index_mappings = await _get_test_index_mappings(dummy_opensearch_db)
    # Ensure the mappings are different to the expected ones
    assert index_mappings != {"properties": dummy_opensearch_db.fields}


async def _get_test_index_mappings(dummy_opensearch_db: DummyOSDB):
    document_id = 1
    index_name = dummy_opensearch_db.index_name(document_id)

    # At this point the index should not exist yet
    with pytest.raises(opensearchpy.exceptions.NotFoundError):
        await dummy_opensearch_db.client.indices.get_mapping(index_name)

    # Insert document which will automatically create the index based on the template
    await dummy_opensearch_db.upsert(document_id, DUMMY_DOCUMENT)

    # Ensure the result looks as expected and return the mappings
    index_mapping = await dummy_opensearch_db.client.indices.get_mapping(index_name)
    assert list(index_mapping) == [index_name]
    assert list(index_mapping[index_name]) == ["mappings"]
    return index_mapping[index_name]["mappings"]



from __future__ import annotations

__all__ = ("AuthDB", "JobDB", "JobLoggingDB", "SandboxMetadataDB", "TaskQueueDB")

from .auth.db import AuthDB
from .jobs.db import JobDB, JobLoggingDB, TaskQueueDB
from .sandbox_metadata.db import SandboxMetadataDB



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import pytest

from diracx.core.exceptions import InvalidQueryError
from diracx.testing.osdb import DummyOSDB

DOC1 = {
    "DateField": datetime.now(tz=timezone.utc),
    "IntField": 1234,
    "KeywordField0": "a",
    "KeywordField1": "keyword1",
    "KeywordField2": "keyword one",
    "TextField": "text value",
    "UnknownField": "unknown field 1",
}
DOC2 = {
    "DateField": datetime.now(tz=timezone.utc) - timedelta(days=1, minutes=34),
    "IntField": 679,
    "KeywordField0": "c",
    "KeywordField1": "keyword1",
    "KeywordField2": "keyword two",
    "TextField": "another text value",
    "UnknownField": "unknown field 2",
}
DOC3 = {
    "DateField": datetime.now(tz=timezone.utc) - timedelta(days=1),
    "IntField": 42,
    "KeywordField0": "b",
    "KeywordField1": "keyword2",
    "KeywordField2": "keyword two",
    "TextField": "yet another text value",
}


@pytest.fixture()
async def prefilled_db(dummy_opensearch_db: DummyOSDB):
    """Fill the database with dummy records for testing."""
    await dummy_opensearch_db.upsert(798811211, DOC1)
    await dummy_opensearch_db.upsert(998811211, DOC2)
    await dummy_opensearch_db.upsert(798811212, DOC3)

    # Force a refresh to make sure the documents are available
    await dummy_opensearch_db.client.indices.refresh(
        index=f"{dummy_opensearch_db.index_prefix}*"
    )

    yield dummy_opensearch_db


async def test_specified_parameters(prefilled_db: DummyOSDB):
    results = await prefilled_db.search(None, [], [])
    assert len(results) == 3
    assert DOC1 in results and DOC2 in results and DOC3 in results

    results = await prefilled_db.search([], [], [])
    assert len(results) == 3
    assert DOC1 in results and DOC2 in results and DOC3 in results

    results = await prefilled_db.search(["IntField"], [], [])
    expected_results = []
    for doc in [DOC1, DOC2, DOC3]:
        expected_doc = {key: doc[key] for key in {"IntField"}}
        # Ensure the document is not already in the list
        # If it is the all() check below no longer makes sense
        assert expected_doc not in expected_results
        expected_results.append(expected_doc)
    assert len(results) == len(expected_results)
    assert all(result in expected_results for result in results)

    results = await prefilled_db.search(["IntField", "UnknownField"], [], [])
    expected_results = [
        {"IntField": DOC1["IntField"], "UnknownField": DOC1["UnknownField"]},
        {"IntField": DOC2["IntField"], "UnknownField": DOC2["UnknownField"]},
        {"IntField": DOC3["IntField"]},
    ]
    assert len(results) == len(expected_results)
    assert all(result in expected_results for result in results)


async def test_pagination_asc(prefilled_db: DummyOSDB):
    sort = [{"parameter": "IntField", "direction": "asc"}]

    results = await prefilled_db.search(None, [], sort)
    assert results == [DOC3, DOC2, DOC1]

    # Pagination has no effect if a specific page isn't requested
    results = await prefilled_db.search(None, [], sort, per_page=2)
    assert results == [DOC3, DOC2, DOC1]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=1)
    assert results == [DOC3, DOC2]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=2)
    assert results == [DOC1]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=3)
    assert results == []

    results = await prefilled_db.search(None, [], sort, per_page=1, page=1)
    assert results == [DOC3]

    results = await prefilled_db.search(None, [], sort, per_page=1, page=2)
    assert results == [DOC2]

    results = await prefilled_db.search(None, [], sort, per_page=1, page=3)
    assert results == [DOC1]

    results = await prefilled_db.search(None, [], sort, per_page=1, page=4)
    assert results == []


async def test_pagination_desc(prefilled_db: DummyOSDB):
    sort = [{"parameter": "IntField", "direction": "desc"}]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=1)
    assert results == [DOC1, DOC2]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=2)
    assert results == [DOC3]


async def test_eq_filter_long(prefilled_db: DummyOSDB):
    part = {"parameter": "IntField", "operator": "eq"}

    # Search for an ID which doesn't exist
    results = await prefilled_db.search(None, [part | {"value": "78"}], [])
    assert results == []

    # Check the DB contains what we expect when not filtering
    results = await prefilled_db.search(None, [], [])
    assert len(results) == 3
    assert DOC1 in results
    assert DOC2 in results
    assert DOC3 in results

    # Search separately for the two documents which do exist
    results = await prefilled_db.search(None, [part | {"value": "1234"}], [])
    assert results == [DOC1]
    results = await prefilled_db.search(None, [part | {"value": "679"}], [])
    assert results == [DOC2]
    results = await prefilled_db.search(None, [part | {"value": "42"}], [])
    assert results == [DOC3]


async def test_operators_long(prefilled_db: DummyOSDB):
    part = {"parameter": "IntField"}

    query = part | {"operator": "neq", "value": "1234"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC2["IntField"], DOC3["IntField"]}

    query = part | {"operator": "in", "values": ["1234", "42"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"], DOC3["IntField"]}

    query = part | {"operator": "lt", "value": "1234"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC2["IntField"], DOC3["IntField"]}

    query = part | {"operator": "lt", "value": "679"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC3["IntField"]}

    query = part | {"operator": "gt", "value": "1234"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    query = part | {"operator": "lt", "value": "42"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()


async def test_operators_date(prefilled_db: DummyOSDB):
    part = {"parameter": "DateField"}

    query = part | {"operator": "eq", "value": DOC3["DateField"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC3["IntField"]}

    query = part | {"operator": "neq", "value": DOC2["DateField"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"], DOC3["IntField"]}

    doc1_time = DOC1["DateField"].strftime("%Y-%m-%dT%H:%M")
    doc2_time = DOC2["DateField"].strftime("%Y-%m-%dT%H:%M")
    doc3_time = DOC3["DateField"].strftime("%Y-%m-%dT%H:%M")

    query = part | {"operator": "in", "values": [doc1_time, doc2_time]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"], DOC2["IntField"]}

    query = part | {"operator": "lt", "value": doc1_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC2["IntField"], DOC3["IntField"]}

    query = part | {"operator": "lt", "value": doc3_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC2["IntField"]}

    query = part | {"operator": "lt", "value": doc2_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    query = part | {"operator": "gt", "value": doc1_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    query = part | {"operator": "gt", "value": doc3_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"]}

    query = part | {"operator": "gt", "value": doc2_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"], DOC3["IntField"]}


@pytest.mark.parametrize(
    "date_format",
    [
        "%Y-%m-%d",
        "%Y-%m-%dT%H",
        "%Y-%m-%dT%H:%M",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%fZ",
    ],
)
async def test_operators_date_partial_doc1(prefilled_db: DummyOSDB, date_format: str):
    """Search by datetime without specifying an exact match

    The parameterized date_format argument should match DOC1 but not DOC2 or DOC3.
    """
    formatted_date = DOC1["DateField"].strftime(date_format)

    query = {"parameter": "DateField", "operator": "eq", "value": formatted_date}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"]}

    query = {"parameter": "DateField", "operator": "neq", "value": formatted_date}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC2["IntField"], DOC3["IntField"]}


async def test_operators_keyword(prefilled_db: DummyOSDB):
    part = {"parameter": "KeywordField1"}

    query = part | {"operator": "eq", "value": DOC1["KeywordField1"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"], DOC2["IntField"]}

    query = part | {"operator": "neq", "value": DOC1["KeywordField1"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC3["IntField"]}

    part = {"parameter": "KeywordField0"}

    query = part | {
        "operator": "in",
        "values": [DOC1["KeywordField0"], DOC3["KeywordField0"]],
    }
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {DOC1["IntField"], DOC3["IntField"]}

    query = part | {"operator": "in", "values": ["missing"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    with pytest.raises(InvalidQueryError):
        query = part | {"operator": "lt", "value": "a"}
        await prefilled_db.search(["IntField"], [query], [])

    with pytest.raises(InvalidQueryError):
        query = part | {"operator": "gt", "value": "a"}
        await prefilled_db.search(["IntField"], [query], [])


async def test_unknown_operator(prefilled_db: DummyOSDB):
    with pytest.raises(InvalidQueryError):
        await prefilled_db.search(
            None, [{"parameter": "IntField", "operator": "unknown"}], []
        )


async def test_unindexed_field(prefilled_db: DummyOSDB):
    with pytest.raises(InvalidQueryError):
        await prefilled_db.search(
            None, [{"parameter": "UnknownField", "eq": "eq", "value": "foobar"}], []
        )


async def test_sort_long(prefilled_db: DummyOSDB):
    results = await prefilled_db.search(
        None, [], [{"parameter": "IntField", "direction": "asc"}]
    )
    assert results == [DOC3, DOC2, DOC1]
    results = await prefilled_db.search(
        None, [], [{"parameter": "IntField", "direction": "desc"}]
    )
    assert results == [DOC1, DOC2, DOC3]


async def test_sort_date(prefilled_db: DummyOSDB):
    results = await prefilled_db.search(
        None, [], [{"parameter": "DateField", "direction": "asc"}]
    )
    assert results == [DOC2, DOC3, DOC1]
    results = await prefilled_db.search(
        None, [], [{"parameter": "DateField", "direction": "desc"}]
    )
    assert results == [DOC1, DOC3, DOC2]


async def test_sort_keyword(prefilled_db: DummyOSDB):
    results = await prefilled_db.search(
        None, [], [{"parameter": "KeywordField0", "direction": "asc"}]
    )
    assert results == [DOC1, DOC3, DOC2]
    results = await prefilled_db.search(
        None, [], [{"parameter": "KeywordField0", "direction": "desc"}]
    )
    assert results == [DOC2, DOC3, DOC1]


async def test_sort_text(prefilled_db: DummyOSDB):
    with pytest.raises(InvalidQueryError):
        await prefilled_db.search(
            None, [], [{"parameter": "TextField", "direction": "asc"}]
        )


async def test_sort_unknown(prefilled_db: DummyOSDB):
    with pytest.raises(InvalidQueryError):
        await prefilled_db.search(
            None, [], [{"parameter": "UnknownField", "direction": "asc"}]
        )


async def test_sort_multiple(prefilled_db: DummyOSDB):
    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "KeywordField1", "direction": "asc"},
            {"parameter": "IntField", "direction": "asc"},
        ],
    )
    assert results == [DOC2, DOC1, DOC3]

    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "KeywordField1", "direction": "asc"},
            {"parameter": "IntField", "direction": "desc"},
        ],
    )
    assert results == [DOC1, DOC2, DOC3]

    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "KeywordField1", "direction": "desc"},
            {"parameter": "IntField", "direction": "asc"},
        ],
    )
    assert results == [DOC3, DOC2, DOC1]

    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "IntField", "direction": "asc"},
            {"parameter": "KeywordField1", "direction": "asc"},
        ],
    )
    assert results == [DOC3, DOC2, DOC1]



from sqlalchemy import (
    BigInteger,
    Boolean,
    Index,
    Integer,
    PrimaryKeyConstraint,
    String,
    UniqueConstraint,
)
from sqlalchemy.orm import declarative_base

from diracx.db.sql.utils import Column, DateNowColumn

Base = declarative_base()


class sb_Owners(Base):
    __tablename__ = "sb_Owners"
    OwnerID = Column(Integer, autoincrement=True)
    Owner = Column(String(32))
    OwnerGroup = Column(String(32))
    VO = Column(String(64))
    __table_args__ = (PrimaryKeyConstraint("OwnerID"),)


class sb_SandBoxes(Base):
    __tablename__ = "sb_SandBoxes"
    SBId = Column(Integer, autoincrement=True)
    OwnerId = Column(Integer)
    SEName = Column(String(64))
    SEPFN = Column(String(512))
    Bytes = Column(BigInteger)
    RegistrationTime = DateNowColumn()
    LastAccessTime = DateNowColumn()
    Assigned = Column(Boolean, default=False)
    __table_args__ = (
        PrimaryKeyConstraint("SBId"),
        Index("OwnerId", OwnerId),
        UniqueConstraint("SEName", "SEPFN", name="Location"),
    )


class sb_EntityMapping(Base):
    __tablename__ = "sb_EntityMapping"
    SBId = Column(Integer)
    EntityId = Column(String(128))
    Type = Column(String(64))
    __table_args__ = (
        PrimaryKeyConstraint("SBId", "EntityId", "Type"),
        Index("SBId", "EntityId"),
        UniqueConstraint("SBId", "EntityId", "Type", name="Mapping"),
    )



from enum import Enum, auto

from sqlalchemy import (
    JSON,
    String,
    Uuid,
)
from sqlalchemy.orm import declarative_base

from diracx.db.sql.utils import Column, DateNowColumn, EnumColumn, NullColumn

USER_CODE_LENGTH = 8

Base = declarative_base()


class FlowStatus(Enum):
    """
    The normal flow is
    PENDING -> READY -> DONE
    Pending is upon insertion
    Ready/Error is set in response to IdP
    Done means the user has been issued the dirac token
    """

    # The flow is ongoing
    PENDING = auto()
    # The user has been authenticated
    # and we are ready to issue a token
    READY = auto()
    # The token has been minted,
    # the user can no longer login with this flow
    DONE = auto()

    # Something went wrong.
    # Most likely the user did not login in the
    # external IdP
    ERROR = auto()


class DeviceFlows(Base):
    __tablename__ = "DeviceFlows"
    user_code = Column(String(USER_CODE_LENGTH), primary_key=True)
    status = EnumColumn(FlowStatus, server_default=FlowStatus.PENDING.name)
    creation_time = DateNowColumn()
    client_id = Column(String(255))
    scope = Column(String(1024))
    audience = Column(String(255))
    device_code = Column(String(128), unique=True)  # hash it ?
    id_token = NullColumn(JSON())


class AuthorizationFlows(Base):
    __tablename__ = "AuthorizationFlows"
    uuid = Column(Uuid(as_uuid=False), primary_key=True)
    status = EnumColumn(FlowStatus, server_default=FlowStatus.PENDING.name)
    client_id = Column(String(255))
    creation_time = DateNowColumn()
    scope = Column(String(1024))
    audience = Column(String(255))
    code_challenge = Column(String(255))
    code_challenge_method = Column(String(8))
    redirect_uri = Column(String(255))
    code = NullColumn(String(255))  # hash it ?
    id_token = NullColumn(JSON())


class RefreshTokenStatus(Enum):
    """
    The normal flow is
    CREATED -> REVOKED

    Note1: There is no EXPIRED status as it can be calculated from a creation time
    Note2: As part of the refresh token rotation mechanism, the revoked token should be retained
    """

    # The token has been created
    # It does not indicate whether the token is valid or not
    CREATED = auto()

    # The token has been revoked
    REVOKED = auto()


class RefreshTokens(Base):
    """Store attributes bound to a refresh token, as well as specific user attributes
    that might be then used to generate access tokens
    """

    __tablename__ = "RefreshTokens"
    # Refresh token attributes
    jti = Column(Uuid(as_uuid=False), primary_key=True)
    status = EnumColumn(
        RefreshTokenStatus, server_default=RefreshTokenStatus.CREATED.name
    )
    creation_time = DateNowColumn()
    scope = Column(String(1024))

    # User attributes bound to the refresh token
    sub = Column(String(1024))
    preferred_username = Column(String(255))



from __future__ import annotations

import pytest
from sqlalchemy.exc import NoResultFound

from diracx.core.exceptions import AuthorizationError
from diracx.db.sql.auth.db import AuthDB

MAX_VALIDITY = 2
EXPIRED = 0


@pytest.fixture
async def auth_db(tmp_path):
    auth_db = AuthDB("sqlite+aiosqlite:///:memory:")
    async with auth_db.engine_context():
        async with auth_db.engine.begin() as conn:
            await conn.run_sync(auth_db.metadata.create_all)
        yield auth_db


async def test_insert_id_token(auth_db: AuthDB):
    # First insert
    async with auth_db as auth_db:
        uuid = await auth_db.insert_authorization_flow(
            "client_id", "scope", "audience", "code_challenge", "S256", "redirect_uri"
        )

    id_token = {"sub": "myIdToken"}

    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            code, redirect_uri = await auth_db.authorization_flow_insert_id_token(
                uuid, id_token, EXPIRED
            )
        code, redirect_uri = await auth_db.authorization_flow_insert_id_token(
            uuid, id_token, MAX_VALIDITY
        )
        assert redirect_uri == "redirect_uri"

    # Cannot add a id_token a second time
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.authorization_flow_insert_id_token(
                uuid, id_token, MAX_VALIDITY
            )

    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.get_authorization_flow(code, EXPIRED)
        res = await auth_db.get_authorization_flow(code, MAX_VALIDITY)
        assert res["id_token"] == id_token

    # Cannot add a id_token after finishing the flow
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.authorization_flow_insert_id_token(
                uuid, id_token, MAX_VALIDITY
            )

    # We shouldn't be able to retrieve it twice
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError, match="already used"):
            res = await auth_db.get_authorization_flow(code, MAX_VALIDITY)


async def test_insert(auth_db: AuthDB):
    # First insert
    async with auth_db as auth_db:
        uuid1 = await auth_db.insert_authorization_flow(
            "client_id", "scope", "audience", "code_challenge", "S256", "redirect_uri"
        )
        uuid2 = await auth_db.insert_authorization_flow(
            "client_id2",
            "scope2",
            "audience2",
            "code_challenge2",
            "S256",
            "redirect_uri2",
        )

    assert uuid1 != uuid2



from __future__ import annotations

import asyncio

import pytest

from diracx.core.exceptions import JobNotFound
from diracx.db.sql.jobs.db import JobDB


@pytest.fixture
async def job_db(tmp_path):
    job_db = JobDB("sqlite+aiosqlite:///:memory:")
    async with job_db.engine_context():
        async with job_db.engine.begin() as conn:
            # set PRAGMA foreign_keys=ON if sqlite
            if job_db._db_url.startswith("sqlite"):
                await conn.exec_driver_sql("PRAGMA foreign_keys=ON")
            await conn.run_sync(job_db.metadata.create_all)
        yield job_db


async def test_some_asyncio_code(job_db):
    async with job_db as job_db:
        result = await job_db.search(["JobID"], [], [])
        assert not result

        result = await asyncio.gather(
            *(
                job_db.insert(
                    f"JDL{i}",
                    "owner",
                    "owner_group",
                    "New",
                    "dfdfds",
                    "lhcb",
                )
                for i in range(100)
            )
        )

    async with job_db as job_db:
        result = await job_db.search(["JobID"], [], [])
        assert result


async def test_set_job_command_invalid_job_id(job_db: JobDB):
    async with job_db as job_db:
        with pytest.raises(JobNotFound):
            await job_db.set_job_command(123456, "test_command")



from __future__ import annotations

import secrets

import pytest
from sqlalchemy.exc import NoResultFound

from diracx.core.exceptions import AuthorizationError, ExpiredFlowError
from diracx.db.sql.auth.db import AuthDB
from diracx.db.sql.auth.schema import USER_CODE_LENGTH

MAX_VALIDITY = 2
EXPIRED = 0


@pytest.fixture
async def auth_db(tmp_path):
    auth_db = AuthDB("sqlite+aiosqlite:///:memory:")
    async with auth_db.engine_context():
        async with auth_db.engine.begin() as conn:
            await conn.run_sync(auth_db.metadata.create_all)
        yield auth_db


async def test_device_user_code_collision(auth_db: AuthDB, monkeypatch):
    monkeypatch.setattr(secrets, "choice", lambda _: "A")

    # First insert should work
    async with auth_db as auth_db:
        code, device = await auth_db.insert_device_flow(
            "client_id", "scope", "audience"
        )
        assert code == "A" * USER_CODE_LENGTH
        assert device

    async with auth_db as auth_db:
        with pytest.raises(NotImplementedError, match="insert new device flow"):
            await auth_db.insert_device_flow("client_id", "scope", "audience")

    monkeypatch.setattr(secrets, "choice", lambda _: "B")

    async with auth_db as auth_db:
        code, device = await auth_db.insert_device_flow(
            "client_id", "scope", "audience"
        )
        assert code == "B" * USER_CODE_LENGTH
        assert device


async def test_device_flow_lookup(auth_db: AuthDB, monkeypatch):
    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.device_flow_validate_user_code("NotInserted", MAX_VALIDITY)

    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.get_device_flow("NotInserted", MAX_VALIDITY)

    # First insert
    async with auth_db as auth_db:
        user_code1, device_code1 = await auth_db.insert_device_flow(
            "client_id1", "scope1", "audience1"
        )
        user_code2, device_code2 = await auth_db.insert_device_flow(
            "client_id2", "scope2", "audience2"
        )

        assert user_code1 != user_code2

    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.device_flow_validate_user_code(user_code1, EXPIRED)

        await auth_db.device_flow_validate_user_code(user_code1, MAX_VALIDITY)

        # Cannot get it with device_code because no id_token
        with pytest.raises(AuthorizationError):
            await auth_db.get_device_flow(device_code1, MAX_VALIDITY)

        await auth_db.device_flow_validate_user_code(user_code2, MAX_VALIDITY)

        # Cannot get it with device_code because no id_token
        with pytest.raises(AuthorizationError):
            await auth_db.get_device_flow(device_code2, MAX_VALIDITY)

    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.device_flow_insert_id_token(
                user_code1, {"token": "mytoken"}, EXPIRED
            )

        await auth_db.device_flow_insert_id_token(
            user_code1, {"token": "mytoken"}, MAX_VALIDITY
        )

        # We should not be able to insert a id_token a second time
        with pytest.raises(AuthorizationError):
            await auth_db.device_flow_insert_id_token(
                user_code1, {"token": "mytoken2"}, MAX_VALIDITY
            )

        with pytest.raises(ExpiredFlowError):
            await auth_db.get_device_flow(device_code1, EXPIRED)

        res = await auth_db.get_device_flow(device_code1, MAX_VALIDITY)
        assert res["user_code"] == user_code1
        assert res["id_token"] == {"token": "mytoken"}

    # cannot get it a second time
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.get_device_flow(device_code1, MAX_VALIDITY)

    # Re-adding a token should not work after it's been minted
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.device_flow_insert_id_token(
                user_code1, {"token": "mytoken"}, MAX_VALIDITY
            )


async def test_device_flow_insert_id_token(auth_db: AuthDB):
    # First insert
    async with auth_db as auth_db:
        user_code, device_code = await auth_db.insert_device_flow(
            "client_id", "scope", "audience"
        )

    # Make sure it exists, and is Pending
    async with auth_db as auth_db:
        await auth_db.device_flow_validate_user_code(user_code, MAX_VALIDITY)

    id_token = {"sub": "myIdToken"}

    async with auth_db as auth_db:
        await auth_db.device_flow_insert_id_token(user_code, id_token, MAX_VALIDITY)

    # The user code has been invalidated
    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.device_flow_validate_user_code(user_code, MAX_VALIDITY)

    async with auth_db as auth_db:
        res = await auth_db.get_device_flow(device_code, MAX_VALIDITY)
        assert res["id_token"] == id_token



from __future__ import annotations

__all__ = ("BaseOSDB",)

import contextlib
import json
import logging
import os
from abc import ABCMeta, abstractmethod
from contextvars import ContextVar
from datetime import datetime
from typing import Any, AsyncIterator, Self

from opensearchpy import AsyncOpenSearch

from diracx.core.exceptions import InvalidQueryError
from diracx.core.extensions import select_from_extension
from diracx.db.exceptions import DBUnavailable

logger = logging.getLogger(__name__)


class OpenSearchDBError(Exception):
    pass


class OpenSearchDBUnavailable(DBUnavailable, OpenSearchDBError):
    pass


class BaseOSDB(metaclass=ABCMeta):
    # TODO: Make metadata an abstract property
    fields: dict
    index_prefix: str

    @abstractmethod
    def index_name(self, doc_id: int) -> str:
        ...

    def __init__(self, connection_kwargs: dict[str, Any]) -> None:
        self._client: AsyncOpenSearch | None = None
        self._connection_kwargs = connection_kwargs
        # We use a ContextVar to make sure that self._conn
        # is specific to each context, and avoid parallel
        # route executions to overlap
        self._conn: ContextVar[bool] = ContextVar("_conn", default=False)

    @classmethod
    def available_implementations(cls, db_name: str) -> list[type[BaseOSDB]]:
        """Return the available implementations of the DB in reverse priority order."""
        db_classes: list[type[BaseOSDB]] = [
            entry_point.load()
            for entry_point in select_from_extension(group="diracx.db.os", name=db_name)
        ]
        if not db_classes:
            raise NotImplementedError(f"Could not find any matches for {db_name=}")
        return db_classes

    @classmethod
    def available_urls(cls) -> dict[str, dict[str, Any]]:
        """Return a dict of available OpenSearch database urls.

        The list of available URLs is determined by environment variables
        prefixed with ``DIRACX_OS_DB_{DB_NAME}``.
        """
        conn_kwargs: dict[str, dict[str, Any]] = {}
        for entry_point in select_from_extension(group="diracx.db.os"):
            db_name = entry_point.name
            var_name = f"DIRACX_OS_DB_{entry_point.name.upper()}"
            if var_name in os.environ:
                try:
                    conn_kwargs[db_name] = json.loads(os.environ[var_name])
                except Exception:
                    logger.error("Error loading connection parameters for %s", db_name)
                    raise
        return conn_kwargs

    @classmethod
    def session(cls) -> Self:
        """This is just a fake method such that the Dependency overwrite has
        a hash to use"""
        raise NotImplementedError("This should never be called")

    @property
    def client(self) -> AsyncOpenSearch:
        """Just a getter for _client, making sure we entered
        the context manager"""
        if self._client is None:
            raise RuntimeError(f"{self.__class__} was used before entering")
        return self._client

    @contextlib.asynccontextmanager
    async def client_context(self) -> AsyncIterator[None]:
        """Context manage to manage the client lifecycle.
        This is called when starting fastapi

        """
        assert self._client is None, "client_context cannot be nested"
        async with AsyncOpenSearch(**self._connection_kwargs) as self._client:
            yield
        self._client = None

    async def ping(self):
        """
        Check whether the connection to the DB is still working.
        We could enable the ``pre_ping`` in the engine, but this would
        be ran at every query.
        """
        if not await self.client.ping():
            raise OpenSearchDBUnavailable(
                f"Failed to connect to {self.__class__.__qualname__}"
            )

    async def __aenter__(self):
        """This is entered on every request.
        At the moment it does nothing, however, we keep it here
        in case we ever want to use OpenSearch equivalent of a transaction
        """
        assert not self._conn.get(), "BaseOSDB context cannot be nested"
        assert self._client is not None, "client_context hasn't been entered"
        self._conn.set(True)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        assert self._conn.get()
        self._client = None
        self._conn.set(False)
        return

    async def create_index_template(self) -> None:
        template_body = {
            "template": {"mappings": {"properties": self.fields}},
            "index_patterns": [f"{self.index_prefix}*"],
        }
        result = await self.client.indices.put_index_template(
            name=self.index_prefix, body=template_body
        )
        assert result["acknowledged"]

    async def upsert(self, doc_id, document) -> None:
        # TODO: Implement properly
        response = await self.client.update(
            index=self.index_name(doc_id),
            id=doc_id,
            body={"doc": document, "doc_as_upsert": True},
            params=dict(retry_on_conflict=10),
        )
        print(f"{response=}")

    async def search(
        self, parameters, search, sorts, *, per_page: int = 100, page: int | None = None
    ) -> list[dict[str, Any]]:
        """Search the database for matching results.

        See the DiracX search API documentation for details.
        """
        body = {}
        if parameters:
            body["_source"] = parameters
        if search:
            body["query"] = apply_search_filters(self.fields, search)
        body["sort"] = []
        for sort in sorts:
            field_name = sort["parameter"]
            field_type = self.fields.get(field_name, {}).get("type")
            require_type("sort", field_name, field_type, {"keyword", "long", "date"})
            body["sort"].append({field_name: {"order": sort["direction"]}})

        params = {}
        if page is not None:
            params["from"] = (page - 1) * per_page
            params["size"] = per_page

        response = await self.client.search(
            body=body, params=params, index=f"{self.index_prefix}*"
        )
        hits = [hit["_source"] for hit in response["hits"]["hits"]]

        # Dates are returned as strings, convert them to Python datetimes
        for hit in hits:
            for field_name in hit:
                if field_name not in self.fields:
                    continue
                if self.fields[field_name]["type"] == "date":
                    hit[field_name] = datetime.strptime(
                        hit[field_name], "%Y-%m-%dT%H:%M:%S.%f%z"
                    )

        return hits


def require_type(operator, field_name, field_type, allowed_types):
    if field_type not in allowed_types:
        raise InvalidQueryError(
            f"Cannot apply {operator} to {field_name} ({field_type=}, {allowed_types=})"
        )


def apply_search_filters(db_fields, search):
    """Build an OpenSearch query from the given DiracX search parameters.

    If the searched parameters cannot be efficiently translated to a query for
    OpenSearch an InvalidQueryError exception is raised.
    """
    result = {
        "must": [],
        "must_not": [],
    }
    for query in search:
        field_name = query["parameter"]
        field_type = db_fields.get(field_name, {}).get("type")
        if field_type is None:
            raise InvalidQueryError(
                f"Field {field_name} is not included in the index mapping"
            )

        match operator := query["operator"]:
            case "eq":
                require_type(
                    operator, field_name, field_type, {"keyword", "long", "date"}
                )
                result["must"].append({"term": {field_name: {"value": query["value"]}}})
            case "neq":
                require_type(
                    operator, field_name, field_type, {"keyword", "long", "date"}
                )
                result["must_not"].append(
                    {"term": {field_name: {"value": query["value"]}}}
                )
            case "gt":
                require_type(operator, field_name, field_type, {"long", "date"})
                result["must"].append({"range": {field_name: {"gt": query["value"]}}})
            case "lt":
                require_type(operator, field_name, field_type, {"long", "date"})
                result["must"].append({"range": {field_name: {"lt": query["value"]}}})
            case "in":
                require_type(
                    operator, field_name, field_type, {"keyword", "long", "date"}
                )
                result["must"].append({"terms": {field_name: query["values"]}})
            # TODO: Implement like and ilike
            # If the pattern is a simple "col like 'abc%'", we can use a prefix query
            # Else we need to use a wildcard query where we replace % with * and _ with ?
            # This should also need to handle escaping of %/_/*/?
            case _:
                raise InvalidQueryError(f"Unknown filter {query=}")

    return {"bool": result}



from datetime import datetime, timezone

import pytest

from diracx.core.models import JobStatus
from diracx.db.sql import JobLoggingDB


@pytest.fixture
async def job_logging_db():
    job_logging_db = JobLoggingDB("sqlite+aiosqlite:///:memory:")
    async with job_logging_db.engine_context():
        async with job_logging_db.engine.begin() as conn:
            await conn.run_sync(job_logging_db.metadata.create_all)
        yield job_logging_db


async def test_insert_record(job_logging_db: JobLoggingDB):
    async with job_logging_db as job_logging_db:
        # Arrange
        date = datetime.now(timezone.utc)

        # Act
        await job_logging_db.insert_record(
            1,
            status=JobStatus.RECEIVED,
            minor_status="minor_status",
            application_status="application_status",
            date=date,
            source="pytest",
        )

        # Assert
        res = await job_logging_db.get_records(1)

        assert len(res) == 1
        assert res[0].Status == JobStatus.RECEIVED.value
        assert res[0].MinorStatus == "minor_status"
        assert res[0].ApplicationStatus == "application_status"
        assert res[0].StatusTime == date
        assert res[0].Source == "pytest"



from __future__ import annotations

__all__ = ("utcnow", "Column", "NullColumn", "DateNowColumn", "BaseSQLDB")

import contextlib
import logging
import os
from abc import ABCMeta
from contextvars import ContextVar
from datetime import datetime, timedelta, timezone
from functools import partial
from typing import TYPE_CHECKING, AsyncIterator, Self, cast

from pydantic import parse_obj_as
from sqlalchemy import Column as RawColumn
from sqlalchemy import DateTime, Enum, MetaData, select
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression

from diracx.core.exceptions import InvalidQueryError
from diracx.core.extensions import select_from_extension
from diracx.core.settings import SqlalchemyDsn
from diracx.db.exceptions import DBUnavailable

if TYPE_CHECKING:
    from sqlalchemy.types import TypeEngine

logger = logging.getLogger(__name__)


class utcnow(expression.FunctionElement):
    type: TypeEngine = DateTime()
    inherit_cache: bool = True


@compiles(utcnow, "postgresql")
def pg_utcnow(element, compiler, **kw) -> str:
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(utcnow, "mssql")
def ms_utcnow(element, compiler, **kw) -> str:
    return "GETUTCDATE()"


@compiles(utcnow, "mysql")
def mysql_utcnow(element, compiler, **kw) -> str:
    return "(UTC_TIMESTAMP)"


@compiles(utcnow, "sqlite")
def sqlite_utcnow(element, compiler, **kw) -> str:
    return "DATETIME('now')"


def substract_date(**kwargs: float) -> datetime:
    return datetime.now(tz=timezone.utc) - timedelta(**kwargs)


Column: partial[RawColumn] = partial(RawColumn, nullable=False)
NullColumn: partial[RawColumn] = partial(RawColumn, nullable=True)
DateNowColumn = partial(Column, DateTime(timezone=True), server_default=utcnow())


def EnumColumn(enum_type, **kwargs):
    return Column(Enum(enum_type, native_enum=False, length=16), **kwargs)


class SQLDBError(Exception):
    pass


class SQLDBUnavailable(DBUnavailable, SQLDBError):
    """Used whenever we encounter a problem with the B connection"""


class BaseSQLDB(metaclass=ABCMeta):
    """This should be the base class of all the DiracX DBs"""

    # engine: AsyncEngine
    # TODO: Make metadata an abstract property
    metadata: MetaData

    def __init__(self, db_url: str) -> None:
        # We use a ContextVar to make sure that self._conn
        # is specific to each context, and avoid parallel
        # route executions to overlap
        self._conn: ContextVar[AsyncConnection | None] = ContextVar(
            "_conn", default=None
        )
        self._db_url = db_url
        self._engine: AsyncEngine | None = None

    @classmethod
    def available_implementations(cls, db_name: str) -> list[type[BaseSQLDB]]:
        """Return the available implementations of the DB in reverse priority order."""
        db_classes: list[type[BaseSQLDB]] = [
            entry_point.load()
            for entry_point in select_from_extension(
                group="diracx.db.sql", name=db_name
            )
        ]
        if not db_classes:
            raise NotImplementedError(f"Could not find any matches for {db_name=}")
        return db_classes

    @classmethod
    def available_urls(cls) -> dict[str, str]:
        """Return a dict of available database urls.

        The list of available URLs is determined by environment variables
        prefixed with ``DIRACX_DB_URL_{DB_NAME}``.
        """
        db_urls: dict[str, str] = {}
        for entry_point in select_from_extension(group="diracx.db.sql"):
            db_name = entry_point.name
            var_name = f"DIRACX_DB_URL_{entry_point.name.upper()}"
            if var_name in os.environ:
                try:
                    db_url = os.environ[var_name]
                    if db_url == "sqlite+aiosqlite:///:memory:":
                        db_urls[db_name] = db_url
                    else:
                        db_urls[db_name] = parse_obj_as(SqlalchemyDsn, db_url)
                except Exception:
                    logger.error("Error loading URL for %s", db_name)
                    raise
        return db_urls

    @classmethod
    def transaction(cls) -> Self:
        raise NotImplementedError("This should never be called")

    @property
    def engine(self) -> AsyncEngine:
        """The engine to use for database operations.

        It is normally not necessary to use the engine directly,
        unless you are doing something special, like writing a
        test fixture that gives you a db.


        Requires that the engine_context has been entered.

        """
        assert self._engine is not None, "engine_context must be entered"
        return self._engine

    @contextlib.asynccontextmanager
    async def engine_context(self) -> AsyncIterator[None]:
        """Context manage to manage the engine lifecycle.
        This is called once at the application startup
        (see ``lifetime_functions``)
        """
        assert self._engine is None, "engine_context cannot be nested"

        # Set the pool_recycle to 30mn
        # That should prevent the problem of MySQL expiring connection
        # after 60mn by default
        engine = create_async_engine(self._db_url, pool_recycle=60 * 30)
        self._engine = engine

        yield

        self._engine = None
        await engine.dispose()

    @property
    def conn(self) -> AsyncConnection:
        if self._conn.get() is None:
            raise RuntimeError(f"{self.__class__} was used before entering")
        return cast(AsyncConnection, self._conn.get())

    async def __aenter__(self) -> Self:
        """
        Create a connection.
        This is called by the Dependency mechanism (see ``db_transaction``),
        It will create a new connection/transaction for each route call.
        """
        assert self._conn.get() is None, "BaseSQLDB context cannot be nested"
        try:
            self._conn.set(await self.engine.connect().__aenter__())
        except Exception as e:
            raise SQLDBUnavailable("Cannot connect to DB") from e

        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
        This is called when exciting a route.
        If there was no exception, the changes in the DB are committed.
        Otherwise, they are rollbacked.
        """
        if exc_type is None:
            await self._conn.get().commit()
        await self._conn.get().__aexit__(exc_type, exc, tb)
        self._conn.set(None)

    async def ping(self):
        """
        Check whether the connection to the DB is still working.
        We could enable the ``pre_ping`` in the engine, but this would
        be ran at every query.
        """
        try:
            await self.conn.scalar(select(1))
        except OperationalError as e:
            raise SQLDBUnavailable("Cannot ping the DB") from e


def apply_search_filters(table, stmt, search):
    # Apply any filters
    for query in search:
        column = table.columns[query["parameter"]]
        if query["operator"] == "eq":
            expr = column == query["value"]
        elif query["operator"] == "neq":
            expr = column != query["value"]
        elif query["operator"] == "gt":
            expr = column > query["value"]
        elif query["operator"] == "lt":
            expr = column < query["value"]
        elif query["operator"] == "in":
            expr = column.in_(query["values"])
        elif query["operator"] in "like":
            expr = column.like(query["value"])
        elif query["operator"] in "ilike":
            expr = column.ilike(query["value"])
        else:
            raise InvalidQueryError(f"Unknown filter {query=}")
        stmt = stmt.where(expr)
    return stmt



from __future__ import annotations

import secrets
from datetime import datetime
from uuid import uuid4

from sqlalchemy import insert, select, update
from sqlalchemy.exc import IntegrityError, NoResultFound

from diracx.core.exceptions import (
    AuthorizationError,
    ExpiredFlowError,
    PendingAuthorizationError,
)
from diracx.db.sql.utils import BaseSQLDB, substract_date

from .schema import (
    AuthorizationFlows,
    DeviceFlows,
    FlowStatus,
    RefreshTokens,
    RefreshTokenStatus,
)
from .schema import Base as AuthDBBase

# https://datatracker.ietf.org/doc/html/rfc8628#section-6.1
USER_CODE_ALPHABET = "BCDFGHJKLMNPQRSTVWXZ"
MAX_RETRY = 5


class AuthDB(BaseSQLDB):
    metadata = AuthDBBase.metadata

    async def device_flow_validate_user_code(
        self, user_code: str, max_validity: int
    ) -> str:
        """Validate that the user_code can be used (Pending status, not expired)

        Returns the scope field for the given user_code

        :raises:
            NoResultFound if no such user code currently Pending
        """
        stmt = select(DeviceFlows.scope).where(
            DeviceFlows.user_code == user_code,
            DeviceFlows.status == FlowStatus.PENDING,
            DeviceFlows.creation_time > substract_date(seconds=max_validity),
        )

        return (await self.conn.execute(stmt)).scalar_one()

    async def get_device_flow(self, device_code: str, max_validity: int):
        """
        :raises: NoResultFound
        """
        # The with_for_update
        # prevents that the token is retrieved
        # multiple time concurrently
        stmt = select(
            DeviceFlows,
            (DeviceFlows.creation_time < substract_date(seconds=max_validity)).label(
                "is_expired"
            ),
        ).with_for_update()
        stmt = stmt.where(
            DeviceFlows.device_code == device_code,
        )
        res = dict((await self.conn.execute(stmt)).one()._mapping)

        if res["is_expired"]:
            raise ExpiredFlowError()

        if res["status"] == FlowStatus.READY:
            # Update the status to Done before returning
            await self.conn.execute(
                update(DeviceFlows)
                .where(DeviceFlows.device_code == device_code)
                .values(status=FlowStatus.DONE)
            )
            return res

        if res["status"] == FlowStatus.DONE:
            raise AuthorizationError("Code was already used")

        if res["status"] == FlowStatus.PENDING:
            raise PendingAuthorizationError()

        raise AuthorizationError("Bad state in device flow")

    async def device_flow_insert_id_token(
        self, user_code: str, id_token: dict[str, str], max_validity: int
    ) -> None:
        """
        :raises: AuthorizationError if no such code or status not pending
        """
        stmt = update(DeviceFlows)
        stmt = stmt.where(
            DeviceFlows.user_code == user_code,
            DeviceFlows.status == FlowStatus.PENDING,
            DeviceFlows.creation_time > substract_date(seconds=max_validity),
        )
        stmt = stmt.values(id_token=id_token, status=FlowStatus.READY)
        res = await self.conn.execute(stmt)
        if res.rowcount != 1:
            raise AuthorizationError(
                f"{res.rowcount} rows matched user_code {user_code}"
            )

    async def insert_device_flow(
        self,
        client_id: str,
        scope: str,
        audience: str,
    ) -> tuple[str, str]:
        # Because the user_code might be short, there is a risk of conflicts
        # This is why we retry multiple times
        for _ in range(MAX_RETRY):
            user_code = "".join(
                secrets.choice(USER_CODE_ALPHABET)
                for _ in range(DeviceFlows.user_code.type.length)  # type: ignore
            )
            # user_code = "2QRKPY"
            device_code = secrets.token_urlsafe()
            stmt = insert(DeviceFlows).values(
                client_id=client_id,
                scope=scope,
                audience=audience,
                user_code=user_code,
                device_code=device_code,
            )
            try:
                await self.conn.execute(stmt)

            except IntegrityError:
                continue

            return user_code, device_code
        raise NotImplementedError(
            f"Could not insert new device flow after {MAX_RETRY} retries"
        )

    async def insert_authorization_flow(
        self,
        client_id: str,
        scope: str,
        audience: str,
        code_challenge: str,
        code_challenge_method: str,
        redirect_uri: str,
    ) -> str:
        uuid = str(uuid4())

        stmt = insert(AuthorizationFlows).values(
            uuid=uuid,
            client_id=client_id,
            scope=scope,
            audience=audience,
            code_challenge=code_challenge,
            code_challenge_method=code_challenge_method,
            redirect_uri=redirect_uri,
        )

        await self.conn.execute(stmt)

        return uuid

    async def authorization_flow_insert_id_token(
        self, uuid: str, id_token: dict[str, str], max_validity: int
    ) -> tuple[str, str]:
        """
        returns code, redirect_uri
        :raises: AuthorizationError if no such uuid or status not pending
        """

        code = secrets.token_urlsafe()
        stmt = update(AuthorizationFlows)

        stmt = stmt.where(
            AuthorizationFlows.uuid == uuid,
            AuthorizationFlows.status == FlowStatus.PENDING,
            AuthorizationFlows.creation_time > substract_date(seconds=max_validity),
        )

        stmt = stmt.values(id_token=id_token, code=code, status=FlowStatus.READY)
        res = await self.conn.execute(stmt)

        if res.rowcount != 1:
            raise AuthorizationError(f"{res.rowcount} rows matched uuid {uuid}")

        stmt = select(AuthorizationFlows.code, AuthorizationFlows.redirect_uri)
        stmt = stmt.where(AuthorizationFlows.uuid == uuid)
        row = (await self.conn.execute(stmt)).one()
        return row.code, row.redirect_uri

    async def get_authorization_flow(self, code: str, max_validity: int):
        # The with_for_update
        # prevents that the token is retrieved
        # multiple time concurrently
        stmt = select(AuthorizationFlows).with_for_update()
        stmt = stmt.where(
            AuthorizationFlows.code == code,
            AuthorizationFlows.creation_time > substract_date(seconds=max_validity),
        )

        res = dict((await self.conn.execute(stmt)).one()._mapping)

        if res["status"] == FlowStatus.READY:
            # Update the status to Done before returning
            await self.conn.execute(
                update(AuthorizationFlows)
                .where(AuthorizationFlows.code == code)
                .values(status=FlowStatus.DONE)
            )

            return res

        if res["status"] == FlowStatus.DONE:
            raise AuthorizationError("Code was already used")

        raise AuthorizationError("Bad state in authorization flow")

    async def insert_refresh_token(
        self,
        subject: str,
        preferred_username: str,
        scope: str,
    ) -> tuple[str, datetime]:
        """
        Insert a refresh token in the DB as well as user attributes
        required to generate access tokens.
        """
        # Generate a JWT ID
        jti = str(uuid4())

        # Insert values into the DB
        stmt = insert(RefreshTokens).values(
            jti=jti,
            sub=subject,
            preferred_username=preferred_username,
            scope=scope,
        )
        await self.conn.execute(stmt)

        # Get the creation time of the new tuple: generated by the insert operation
        stmt = select(RefreshTokens.creation_time)
        stmt = stmt.where(RefreshTokens.jti == jti)
        row = (await self.conn.execute(stmt)).one()

        # Return the JWT ID and the creation time
        return jti, row.creation_time

    async def get_refresh_token(self, jti: str) -> dict:
        """
        Get refresh token details bound to a given JWT ID
        """
        # The with_for_update
        # prevents that the token is retrieved
        # multiple time concurrently
        stmt = select(RefreshTokens).with_for_update()
        stmt = stmt.where(
            RefreshTokens.jti == jti,
        )
        try:
            res = dict((await self.conn.execute(stmt)).one()._mapping)
        except NoResultFound:
            return {}

        return res

    async def get_user_refresh_tokens(self, subject: str | None = None) -> list[dict]:
        """Get a list of refresh token details based on a subject ID (not revoked)"""
        # Get a list of refresh tokens
        stmt = select(RefreshTokens).with_for_update()

        if subject:
            stmt = stmt.where(
                RefreshTokens.sub == subject,
                RefreshTokens.status != RefreshTokenStatus.REVOKED,
            )

        res = (await self.conn.execute(stmt)).all()

        # Convert the results into dict
        refresh_tokens = []
        for refresh_token in res:
            refresh_tokens.append(dict(refresh_token._mapping))

        return refresh_tokens

    async def revoke_refresh_token(self, jti: str):
        """Revoke a token given by its JWT ID"""
        await self.conn.execute(
            update(RefreshTokens)
            .where(RefreshTokens.jti == jti)
            .values(status=RefreshTokenStatus.REVOKED)
        )

    async def revoke_user_refresh_tokens(self, subject):
        """Revoke all the refresh tokens belonging to a user (subject ID)"""
        await self.conn.execute(
            update(RefreshTokens)
            .where(RefreshTokens.sub == subject)
            .values(status=RefreshTokenStatus.REVOKED)
        )



from __future__ import annotations

import argparse
import asyncio
import logging

logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(required=True)

    init_sql_parser = subparsers.add_parser(
        "init-sql", help="Initialise schema for SQL databases"
    )
    init_sql_parser.set_defaults(func=init_sql)

    init_os_parser = subparsers.add_parser(
        "init-os", help="Initialise schema for OpenSearch databases"
    )
    init_os_parser.set_defaults(func=init_os)

    args = parser.parse_args()
    logger.setLevel(logging.INFO)
    asyncio.run(args.func())


async def init_sql():
    logger.info("Initialising SQL databases")
    from diracx.db.sql.utils import BaseSQLDB

    for db_name, db_url in BaseSQLDB.available_urls().items():
        logger.info("Initialising %s", db_name)
        db = BaseSQLDB.available_implementations(db_name)[0](db_url)
        async with db.engine_context():
            async with db.engine.begin() as conn:
                # set PRAGMA foreign_keys=ON if sqlite
                if db._db_url.startswith("sqlite"):
                    await conn.exec_driver_sql("PRAGMA foreign_keys=ON")
                await conn.run_sync(db.metadata.create_all)


async def init_os():
    logger.info("Initialising OpenSearch databases")
    from diracx.db.os.utils import BaseOSDB

    for db_name, db_url in BaseOSDB.available_urls().items():
        logger.info("Initialising %s", db_name)
        db = BaseOSDB.available_implementations(db_name)[0](db_url)
        async with db.client_context():
            await db.create_index_template()


if __name__ == "__main__":
    parse_args()



import sqlalchemy.types as types
from sqlalchemy import (
    BigInteger,
    Boolean,
    DateTime,
    Enum,
    Float,
    ForeignKey,
    Index,
    Integer,
    Numeric,
    PrimaryKeyConstraint,
    String,
    Text,
)
from sqlalchemy.orm import declarative_base

from ..utils import Column, DateNowColumn, NullColumn

JobDBBase = declarative_base()
JobLoggingDBBase = declarative_base()
TaskQueueDBBase = declarative_base()


class EnumBackedBool(types.TypeDecorator):
    """Maps a ``EnumBackedBool()`` column to True/False in Python"""

    impl = types.Enum
    cache_ok: bool = True

    def __init__(self) -> None:
        super().__init__("True", "False")

    def process_bind_param(self, value, dialect) -> str:
        if value is True:
            return "True"
        elif value is False:
            return "False"
        else:
            raise NotImplementedError(value, dialect)

    def process_result_value(self, value, dialect) -> bool:
        if value == "True":
            return True
        elif value == "False":
            return False
        else:
            raise NotImplementedError(f"Unknown {value=}")


class Jobs(JobDBBase):
    __tablename__ = "Jobs"

    JobID = Column(
        "JobID",
        Integer,
        ForeignKey("JobJDLs.JobID", ondelete="CASCADE"),
        primary_key=True,
        default=0,
    )
    JobType = Column("JobType", String(32), default="user")
    JobGroup = Column("JobGroup", String(32), default="00000000")
    Site = Column("Site", String(100), default="ANY")
    JobName = Column("JobName", String(128), default="Unknown")
    Owner = Column("Owner", String(64), default="Unknown")
    OwnerGroup = Column("OwnerGroup", String(128), default="Unknown")
    VO = Column("VO", String(32))
    SubmissionTime = NullColumn("SubmissionTime", DateTime)
    RescheduleTime = NullColumn("RescheduleTime", DateTime)
    LastUpdateTime = NullColumn("LastUpdateTime", DateTime)
    StartExecTime = NullColumn("StartExecTime", DateTime)
    HeartBeatTime = NullColumn("HeartBeatTime", DateTime)
    EndExecTime = NullColumn("EndExecTime", DateTime)
    Status = Column("Status", String(32), default="Received")
    MinorStatus = Column("MinorStatus", String(128), default="Unknown")
    ApplicationStatus = Column("ApplicationStatus", String(255), default="Unknown")
    UserPriority = Column("UserPriority", Integer, default=0)
    RescheduleCounter = Column("RescheduleCounter", Integer, default=0)
    VerifiedFlag = Column("VerifiedFlag", EnumBackedBool(), default=False)
    # TODO: Should this be True/False/"Failed"? Or True/False/Null?
    AccountedFlag = Column(
        "AccountedFlag", Enum("True", "False", "Failed"), default="False"
    )

    __table_args__ = (
        Index("JobType", "JobType"),
        Index("JobGroup", "JobGroup"),
        Index("Site", "Site"),
        Index("Owner", "Owner"),
        Index("OwnerGroup", "OwnerGroup"),
        Index("Status", "Status"),
        Index("MinorStatus", "MinorStatus"),
        Index("ApplicationStatus", "ApplicationStatus"),
        Index("StatusSite", "Status", "Site"),
        Index("LastUpdateTime", "LastUpdateTime"),
    )


class JobJDLs(JobDBBase):
    __tablename__ = "JobJDLs"
    JobID = Column(Integer, autoincrement=True, primary_key=True)
    JDL = Column(Text)
    JobRequirements = Column(Text)
    OriginalJDL = Column(Text)


class InputData(JobDBBase):
    __tablename__ = "InputData"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    LFN = Column(String(255), default="", primary_key=True)
    Status = Column(String(32), default="AprioriGood")


class JobParameters(JobDBBase):
    __tablename__ = "JobParameters"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)


class OptimizerParameters(JobDBBase):
    __tablename__ = "OptimizerParameters"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)


class AtticJobParameters(JobDBBase):
    __tablename__ = "AtticJobParameters"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)
    RescheduleCycle = Column(Integer)


class SiteMask(JobDBBase):
    __tablename__ = "SiteMask"
    Site = Column(String(64), primary_key=True)
    Status = Column(String(64))
    LastUpdateTime = Column(DateTime)
    Author = Column(String(255))
    Comment = Column(Text)


class SiteMaskLogging(JobDBBase):
    __tablename__ = "SiteMaskLogging"
    Site = Column(String(64), primary_key=True)
    UpdateTime = Column(DateTime, primary_key=True)
    Status = Column(String(64))
    Author = Column(String(255))
    Comment = Column(Text)


class HeartBeatLoggingInfo(JobDBBase):
    __tablename__ = "HeartBeatLoggingInfo"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Name = Column(String(100), primary_key=True)
    Value = Column(Text)
    HeartBeatTime = Column(DateTime, primary_key=True)


class JobCommands(JobDBBase):
    __tablename__ = "JobCommands"
    JobID = Column(
        Integer, ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    Command = Column(String(100))
    Arguments = Column(String(100))
    Status = Column(String(64), default="Received")
    ReceptionTime = Column(DateTime, primary_key=True)
    ExecutionTime = NullColumn(DateTime)


class LoggingInfo(JobLoggingDBBase):
    __tablename__ = "LoggingInfo"
    JobID = Column(Integer)
    SeqNum = Column(Integer)
    Status = Column(String(32), default="")
    MinorStatus = Column(String(128), default="")
    ApplicationStatus = Column(String(255), default="")
    StatusTime = DateNowColumn()
    # TODO: Check that this corresponds to the DOUBLE(12,3) type in MySQL
    StatusTimeOrder = Column(Numeric(precision=12, scale=3), default=0)
    Source = Column(String(32), default="Unknown", name="StatusSource")
    __table_args__ = (PrimaryKeyConstraint("JobID", "SeqNum"),)


class TaskQueues(TaskQueueDBBase):
    __tablename__ = "tq_TaskQueues"
    TQId = Column(Integer, primary_key=True)
    Owner = Column(String(255), nullable=False)
    OwnerGroup = Column(String(32), nullable=False)
    VO = Column(String(32), nullable=False)
    CPUTime = Column(BigInteger, nullable=False)
    Priority = Column(Float, nullable=False)
    Enabled = Column(Boolean, nullable=False, default=0)
    __table_args__ = (Index("TQOwner", "Owner", "OwnerGroup", "CPUTime"),)


class JobsQueue(TaskQueueDBBase):
    __tablename__ = "tq_Jobs"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    JobId = Column(Integer, primary_key=True)
    Priority = Column(Integer, nullable=False)
    RealPriority = Column(Float, nullable=False)
    __table_args__ = (Index("TaskIndex", "TQId"),)


class SitesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToSites"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("SitesTaskIndex", "TQId"),
        Index("SitesIndex", "Value"),
    )


class GridCEsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToGridCEs"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("GridCEsTaskIndex", "TQId"),
        Index("GridCEsValueIndex", "Value"),
    )


class BannedSitesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToBannedSites"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("BannedSitesTaskIndex", "TQId"),
        Index("BannedSitesValueIndex", "Value"),
    )


class PlatformsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToPlatforms"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("PlatformsTaskIndex", "TQId"),
        Index("PlatformsValueIndex", "Value"),
    )


class JobTypesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToJobTypes"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("JobTypesTaskIndex", "TQId"),
        Index("JobTypesValueIndex", "Value"),
    )


class TagsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToTags"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("TagsTaskIndex", "TQId"),
        Index("TagsValueIndex", "Value"),
    )



from __future__ import annotations

from uuid import UUID

from sqlalchemy import func, insert, select

from diracx.db.sql.utils import BaseSQLDB, apply_search_filters

from .schema import Base as DummyDBBase
from .schema import Cars, Owners


class DummyDB(BaseSQLDB):
    """
    This DummyDB is just to illustrate some important aspect of writing
    DB classes in DiracX.

    It is mostly pure SQLAlchemy, with a few convention

    Document the secrets
    """

    # This needs to be here for the BaseSQLDB to create the engine
    metadata = DummyDBBase.metadata

    async def summary(self, group_by, search) -> list[dict[str, str | int]]:
        columns = [Cars.__table__.columns[x] for x in group_by]

        stmt = select(*columns, func.count(Cars.licensePlate).label("count"))
        stmt = apply_search_filters(Cars.__table__, stmt, search)
        stmt = stmt.group_by(*columns)

        # Execute the query
        return [
            dict(row._mapping)
            async for row in (await self.conn.stream(stmt))
            if row.count > 0  # type: ignore
        ]

    async def insert_owner(self, name: str) -> int:
        stmt = insert(Owners).values(name=name)
        result = await self.conn.execute(stmt)
        # await self.engine.commit()
        return result.lastrowid

    async def insert_car(self, license_plate: UUID, model: str, owner_id: int) -> int:
        stmt = insert(Cars).values(
            licensePlate=license_plate, model=model, ownerID=owner_id
        )

        result = await self.conn.execute(stmt)
        # await self.engine.commit()
        return result.lastrowid



from __future__ import annotations

import sqlalchemy
from sqlalchemy import delete

from diracx.core.models import SandboxInfo, UserInfo
from diracx.db.sql.utils import BaseSQLDB, utcnow

from .schema import Base as SandboxMetadataDBBase
from .schema import sb_EntityMapping, sb_Owners, sb_SandBoxes


class SandboxMetadataDB(BaseSQLDB):
    metadata = SandboxMetadataDBBase.metadata

    async def upsert_owner(self, user: UserInfo) -> int:
        """Get the id of the owner from the database"""
        # TODO: Follow https://github.com/DIRACGrid/diracx/issues/49
        stmt = sqlalchemy.select(sb_Owners.OwnerID).where(
            sb_Owners.Owner == user.preferred_username,
            sb_Owners.OwnerGroup == user.dirac_group,
            sb_Owners.VO == user.vo,
        )
        result = await self.conn.execute(stmt)
        if owner_id := result.scalar_one_or_none():
            return owner_id

        stmt = sqlalchemy.insert(sb_Owners).values(
            Owner=user.preferred_username,
            OwnerGroup=user.dirac_group,
            VO=user.vo,
        )
        result = await self.conn.execute(stmt)
        return result.lastrowid

    @staticmethod
    def get_pfn(bucket_name: str, user: UserInfo, sandbox_info: SandboxInfo) -> str:
        """Get the sandbox's user namespaced and content addressed PFN"""
        parts = [
            "S3",
            bucket_name,
            user.vo,
            user.dirac_group,
            user.preferred_username,
            f"{sandbox_info.checksum_algorithm}:{sandbox_info.checksum}.{sandbox_info.format}",
        ]
        return "/" + "/".join(parts)

    async def insert_sandbox(
        self, se_name: str, user: UserInfo, pfn: str, size: int
    ) -> None:
        """Add a new sandbox in SandboxMetadataDB"""
        # TODO: Follow https://github.com/DIRACGrid/diracx/issues/49
        owner_id = await self.upsert_owner(user)
        stmt = sqlalchemy.insert(sb_SandBoxes).values(
            OwnerId=owner_id,
            SEName=se_name,
            SEPFN=pfn,
            Bytes=size,
            RegistrationTime=utcnow(),
            LastAccessTime=utcnow(),
        )
        try:
            result = await self.conn.execute(stmt)
        except sqlalchemy.exc.IntegrityError:
            await self.update_sandbox_last_access_time(se_name, pfn)
        else:
            assert result.rowcount == 1

    async def update_sandbox_last_access_time(self, se_name: str, pfn: str) -> None:
        stmt = (
            sqlalchemy.update(sb_SandBoxes)
            .where(sb_SandBoxes.SEName == se_name, sb_SandBoxes.SEPFN == pfn)
            .values(LastAccessTime=utcnow())
        )
        result = await self.conn.execute(stmt)
        assert result.rowcount == 1

    async def sandbox_is_assigned(self, se_name: str, pfn: str) -> bool:
        """Checks if a sandbox exists and has been assigned."""
        stmt: sqlalchemy.Executable = sqlalchemy.select(sb_SandBoxes.Assigned).where(
            sb_SandBoxes.SEName == se_name, sb_SandBoxes.SEPFN == pfn
        )
        result = await self.conn.execute(stmt)
        is_assigned = result.scalar_one()
        return is_assigned
        return True

    async def unassign_sandbox_from_jobs(self, job_ids: list[int]):
        """
        Unassign sandbox from jobs
        """
        stmt = delete(sb_EntityMapping).where(
            sb_EntityMapping.EntityId.in_(f"Job:{job_id}" for job_id in job_ids)
        )
        await self.conn.execute(stmt)



# The utils class define some boilerplate types that should be used
# in place of the SQLAlchemy one. Have a look at them
from sqlalchemy import ForeignKey, Integer, String, Uuid
from sqlalchemy.orm import declarative_base

from diracx.db.sql.utils import Column, DateNowColumn

Base = declarative_base()


class Owners(Base):
    __tablename__ = "Owners"
    ownerID = Column(Integer, primary_key=True, autoincrement=True)
    creation_time = DateNowColumn()
    name = Column(String(255))


class Cars(Base):
    __tablename__ = "Cars"
    licensePlate = Column(Uuid(), primary_key=True)
    model = Column(String(255))
    ownerID = Column(Integer, ForeignKey(Owners.ownerID))



from __future__ import annotations

import pytest

from diracx.db.os.utils import OpenSearchDBUnavailable
from diracx.testing.osdb import OPENSEARCH_PORT, DummyOSDB, require_port_availability


async def _ensure_db_unavailable(db: DummyOSDB):
    """Helper function which raises an exception if we manage to connect to the DB."""
    async with db.client_context():
        async with db:
            with pytest.raises(OpenSearchDBUnavailable):
                await db.ping()


async def test_connection(dummy_opensearch_db: DummyOSDB):
    """Ensure we can connect to the OpenSearch database."""
    assert await dummy_opensearch_db.client.ping()


async def test_connection_error_bad_port(opensearch_conn_kwargs):
    """Check the connection behavior when the DB is unavailable.

    This failure mode is emulated by changing the port number.
    """
    require_port_availability(28001)
    assert f":{OPENSEARCH_PORT}" in opensearch_conn_kwargs["hosts"]
    db = DummyOSDB(
        {
            **opensearch_conn_kwargs,
            "hosts": opensearch_conn_kwargs["hosts"].replace(
                f":{OPENSEARCH_PORT}", ":28001"
            ),
        }
    )
    await _ensure_db_unavailable(db)


async def test_connection_error_ssl(opensearch_conn_kwargs):
    """Check the connection behavior when there is an SSL error."""
    db = DummyOSDB({**opensearch_conn_kwargs, "use_ssl": False})
    await _ensure_db_unavailable(db)


async def test_connection_error_certs(opensearch_conn_kwargs):
    """Check the connection behavior when there is an certificate verification error."""
    db = DummyOSDB({**opensearch_conn_kwargs, "verify_certs": True})
    await _ensure_db_unavailable(db)


async def test_connection_error_bad_username(opensearch_conn_kwargs):
    """Check the connection behavior when the username is incorrect."""
    assert "admin:admin" in opensearch_conn_kwargs["hosts"]
    db = DummyOSDB(
        {
            **opensearch_conn_kwargs,
            "hosts": opensearch_conn_kwargs["hosts"].replace(
                "admin:admin", "nobody:admin"
            ),
        }
    )
    await _ensure_db_unavailable(db)


async def test_connection_error_bad_password(opensearch_conn_kwargs):
    """Check the connection behavior when the password is incorrect."""
    assert "admin:admin" in opensearch_conn_kwargs["hosts"]
    db = DummyOSDB(
        {
            **opensearch_conn_kwargs,
            "hosts": opensearch_conn_kwargs["hosts"].replace(
                "admin:admin", "admin:wrong"
            ),
        }
    )
    await _ensure_db_unavailable(db)


async def test_sanity_checks(opensearch_conn_kwargs):
    """Check that the sanity checks are working as expected."""
    db = DummyOSDB(opensearch_conn_kwargs)
    # Check that the client is not available before entering the context manager
    with pytest.raises(RuntimeError):
        await db.ping()

    # It shouldn't be possible to enter the context manager twice
    async with db.client_context():
        async with db:
            await db.ping()



from __future__ import annotations

import logging
import time
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

from sqlalchemy import bindparam, delete, func, insert, select, update
from sqlalchemy.exc import IntegrityError, NoResultFound

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import BindParameter

from diracx.core.exceptions import InvalidQueryError, JobNotFound
from diracx.core.models import (
    JobMinorStatus,
    JobStatus,
    JobStatusReturn,
    LimitedJobStatusReturn,
    ScalarSearchOperator,
    ScalarSearchSpec,
)
from diracx.core.properties import JOB_SHARING, SecurityProperty

from ..utils import BaseSQLDB, apply_search_filters
from .schema import (
    BannedSitesQueue,
    GridCEsQueue,
    InputData,
    JobCommands,
    JobDBBase,
    JobJDLs,
    JobLoggingDBBase,
    Jobs,
    JobsQueue,
    JobTypesQueue,
    LoggingInfo,
    PlatformsQueue,
    SitesQueue,
    TagsQueue,
    TaskQueueDBBase,
    TaskQueues,
)


def _get_columns(table, parameters):
    columns = [x for x in table.columns]
    if parameters:
        if unrecognised_parameters := set(parameters) - set(table.columns.keys()):
            raise InvalidQueryError(
                f"Unrecognised parameters requested {unrecognised_parameters}"
            )
        columns = [c for c in columns if c.name in parameters]
    return columns


class JobDB(BaseSQLDB):
    metadata = JobDBBase.metadata

    # TODO: this is copied from the DIRAC JobDB
    # but is overwriten in LHCbDIRAC, so we need
    # to find a way to make it dynamic
    jdl2DBParameters = ["JobName", "JobType", "JobGroup"]

    # TODO: set maxRescheduling value from CS
    # maxRescheduling = self.getCSOption("MaxRescheduling", 3)
    # For now:
    maxRescheduling = 3

    async def summary(self, group_by, search) -> list[dict[str, str | int]]:
        columns = _get_columns(Jobs.__table__, group_by)

        stmt = select(*columns, func.count(Jobs.JobID).label("count"))
        stmt = apply_search_filters(Jobs.__table__, stmt, search)
        stmt = stmt.group_by(*columns)

        # Execute the query
        return [
            dict(row._mapping)
            async for row in (await self.conn.stream(stmt))
            if row.count > 0  # type: ignore
        ]

    async def search(
        self,
        parameters,
        search,
        sorts,
        *,
        distinct: bool = False,
        per_page: int = 100,
        page: int | None = None,
    ) -> list[dict[str, Any]]:
        # Find which columns to select
        columns = _get_columns(Jobs.__table__, parameters)
        stmt = select(*columns)

        stmt = apply_search_filters(Jobs.__table__, stmt, search)

        # Apply any sort constraints
        for sort in sorts:
            if sort["parameter"] not in Jobs.__table__.columns:
                raise InvalidQueryError(
                    f"Cannot sort by {sort['parameter']}: unknown column"
                )
            column = Jobs.__table__.columns[sort["parameter"]]
            if sort["direction"] == "asc":
                column = column.asc()
            elif sort["direction"] == "desc":
                column = column.desc()
            else:
                raise InvalidQueryError(f"Unknown sort {sort['direction']=}")

        if distinct:
            stmt = stmt.distinct()

        # Apply pagination
        if page:
            raise NotImplementedError("TODO Not yet implemented")

        # Execute the query
        return [dict(row._mapping) async for row in (await self.conn.stream(stmt))]

    async def _insertNewJDL(self, jdl) -> int:
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import compressJDL

        stmt = insert(JobJDLs).values(
            JDL="", JobRequirements="", OriginalJDL=compressJDL(jdl)
        )
        result = await self.conn.execute(stmt)
        # await self.engine.commit()
        return result.lastrowid

    async def _insertJob(self, jobData: dict[str, Any]):
        stmt = insert(Jobs).values(jobData)
        await self.conn.execute(stmt)

    async def _insertInputData(self, job_id: int, lfns: list[str]):
        stmt = insert(InputData).values([{"JobID": job_id, "LFN": lfn} for lfn in lfns])
        await self.conn.execute(stmt)

    async def setJobAttributes(self, job_id, jobData):
        """
        TODO: add myDate and force parameters
        """
        if "Status" in jobData:
            jobData = jobData | {"LastUpdateTime": datetime.now(tz=timezone.utc)}
        stmt = update(Jobs).where(Jobs.JobID == job_id).values(jobData)
        await self.conn.execute(stmt)

    async def _checkAndPrepareJob(
        self,
        jobID,
        class_ad_job,
        class_ad_req,
        owner,
        owner_group,
        job_attrs,
        vo,
    ):
        """
        Check Consistency of Submitted JDL and set some defaults
        Prepare subJDL with Job Requirements
        """
        from DIRAC.Core.Utilities.DErrno import EWMSSUBM, cmpError
        from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import (
            checkAndPrepareJob,
        )

        retVal = checkAndPrepareJob(
            jobID,
            class_ad_job,
            class_ad_req,
            owner,
            owner_group,
            job_attrs,
            vo,
        )

        if not retVal["OK"]:
            if cmpError(retVal, EWMSSUBM):
                await self.setJobAttributes(jobID, job_attrs)

            returnValueOrRaise(retVal)

    async def setJobJDL(self, job_id, jdl):
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import compressJDL

        stmt = (
            update(JobJDLs).where(JobJDLs.JobID == job_id).values(JDL=compressJDL(jdl))
        )
        await self.conn.execute(stmt)

    async def getJobJDL(self, job_id: int, original: bool = False) -> str:
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import extractJDL

        if original:
            stmt = select(JobJDLs.OriginalJDL).where(JobJDLs.JobID == job_id)
        else:
            stmt = select(JobJDLs.JDL).where(JobJDLs.JobID == job_id)

        jdl = (await self.conn.execute(stmt)).scalar_one()
        if jdl:
            jdl = extractJDL(jdl)

        return jdl

    async def insert(
        self,
        jdl,
        owner,
        owner_group,
        initial_status,
        initial_minor_status,
        vo,
    ):
        from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
        from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
        from DIRAC.WorkloadManagementSystem.DB.JobDBUtils import (
            checkAndAddOwner,
            createJDLWithInitialStatus,
            fixJDL,
        )

        job_attrs = {
            "LastUpdateTime": datetime.now(tz=timezone.utc),
            "SubmissionTime": datetime.now(tz=timezone.utc),
            "Owner": owner,
            "OwnerGroup": owner_group,
            "VO": vo,
        }

        jobManifest = returnValueOrRaise(checkAndAddOwner(jdl, owner, owner_group))

        jdl = fixJDL(jdl)

        job_id = await self._insertNewJDL(jdl)

        jobManifest.setOption("JobID", job_id)

        job_attrs["JobID"] = job_id

        # 2.- Check JDL and Prepare DIRAC JDL
        jobJDL = jobManifest.dumpAsJDL()

        # Replace the JobID placeholder if any
        if jobJDL.find("%j") != -1:
            jobJDL = jobJDL.replace("%j", str(job_id))

        class_ad_job = ClassAd(jobJDL)
        class_ad_req = ClassAd("[]")
        if not class_ad_job.isOK():
            job_attrs["Status"] = JobStatus.FAILED

            job_attrs["MinorStatus"] = "Error in JDL syntax"

            await self._insertJob(job_attrs)

            return {
                "JobID": job_id,
                "Status": JobStatus.FAILED,
                "MinorStatus": "Error in JDL syntax",
            }

        class_ad_job.insertAttributeInt("JobID", job_id)

        await self._checkAndPrepareJob(
            job_id,
            class_ad_job,
            class_ad_req,
            owner,
            owner_group,
            job_attrs,
            vo,
        )

        jobJDL = createJDLWithInitialStatus(
            class_ad_job,
            class_ad_req,
            self.jdl2DBParameters,
            job_attrs,
            initial_status,
            initial_minor_status,
            modern=True,
        )

        await self.setJobJDL(job_id, jobJDL)

        # Adding the job in the Jobs table
        await self._insertJob(job_attrs)

        # TODO: check if that is actually true
        if class_ad_job.lookupAttribute("Parameters"):
            raise NotImplementedError("Parameters in the JDL are not supported")

        # Looking for the Input Data
        inputData = []
        if class_ad_job.lookupAttribute("InputData"):
            inputData = class_ad_job.getListFromExpression("InputData")
            lfns = [lfn for lfn in inputData if lfn]
            if lfns:
                await self._insertInputData(job_id, lfns)

        return {
            "JobID": job_id,
            "Status": initial_status,
            "MinorStatus": initial_minor_status,
            "TimeStamp": datetime.now(tz=timezone.utc),
        }

    async def rescheduleJob(self, job_id) -> dict[str, Any]:
        """Reschedule given job"""
        from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
        from DIRAC.Core.Utilities.ReturnValues import SErrorException

        result = await self.search(
            parameters=[
                "Status",
                "MinorStatus",
                "VerifiedFlag",
                "RescheduleCounter",
                "Owner",
                "OwnerGroup",
            ],
            search=[
                ScalarSearchSpec(
                    parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=job_id
                )
            ],
            sorts=[],
        )
        if not result:
            raise ValueError(f"Job {job_id} not found.")

        jobAttrs = result[0]

        if "VerifiedFlag" not in jobAttrs:
            raise ValueError(f"Job {job_id} not found in the system")

        if not jobAttrs["VerifiedFlag"]:
            raise ValueError(
                f"Job {job_id} not Verified: Status {jobAttrs['Status']}, Minor Status: {jobAttrs['MinorStatus']}"
            )

        reschedule_counter = int(jobAttrs["RescheduleCounter"]) + 1

        # TODO: update maxRescheduling:
        # self.maxRescheduling = self.getCSOption("MaxRescheduling", self.maxRescheduling)

        if reschedule_counter > self.maxRescheduling:
            logging.warn(f"Job {job_id}: Maximum number of reschedulings is reached.")
            self.setJobAttributes(
                job_id,
                {
                    "Status": JobStatus.FAILED,
                    "MinorStatus": JobMinorStatus.MAX_RESCHEDULING,
                },
            )
            raise ValueError(
                f"Maximum number of reschedulings is reached: {self.maxRescheduling}"
            )

        new_job_attributes = {"RescheduleCounter": reschedule_counter}

        # TODO: get the job parameters from JobMonitoringClient
        # result = JobMonitoringClient().getJobParameters(jobID)
        # if result["OK"]:
        #     parDict = result["Value"]
        #     for key, value in parDict.get(jobID, {}).items():
        #         result = self.setAtticJobParameter(jobID, key, value, rescheduleCounter - 1)
        #         if not result["OK"]:
        #             break

        # TODO: IF we keep JobParameters and OptimizerParameters: Delete job in those tables.
        # await self.delete_job_parameters(job_id)
        # await self.delete_job_optimizer_parameters(job_id)

        job_jdl = await self.getJobJDL(job_id, original=True)
        if not job_jdl.strip().startswith("["):
            job_jdl = f"[{job_jdl}]"

        classAdJob = ClassAd(job_jdl)
        classAdReq = ClassAd("[]")
        retVal = {}
        retVal["JobID"] = job_id

        classAdJob.insertAttributeInt("JobID", job_id)

        try:
            result = await self._checkAndPrepareJob(
                job_id,
                classAdJob,
                classAdReq,
                jobAttrs["Owner"],
                jobAttrs["OwnerGroup"],
                new_job_attributes,
                classAdJob.getAttributeString("VirtualOrganization"),
            )
        except SErrorException as e:
            raise ValueError(e) from e

        priority = classAdJob.getAttributeInt("Priority")
        if priority is None:
            priority = 0
        jobAttrs["UserPriority"] = priority

        siteList = classAdJob.getListFromExpression("Site")
        if not siteList:
            site = "ANY"
        elif len(siteList) > 1:
            site = "Multiple"
        else:
            site = siteList[0]

        jobAttrs["Site"] = site

        jobAttrs["Status"] = JobStatus.RECEIVED

        jobAttrs["MinorStatus"] = JobMinorStatus.RESCHEDULED

        jobAttrs["ApplicationStatus"] = "Unknown"

        jobAttrs["LastUpdateTime"] = datetime.now(tz=timezone.utc)

        jobAttrs["RescheduleTime"] = datetime.now(tz=timezone.utc)

        reqJDL = classAdReq.asJDL()
        classAdJob.insertAttributeInt("JobRequirements", reqJDL)

        jobJDL = classAdJob.asJDL()

        # Replace the JobID placeholder if any
        jobJDL = jobJDL.replace("%j", str(job_id))

        result = await self.setJobJDL(job_id, jobJDL)

        result = await self.setJobAttributes(job_id, jobAttrs)

        retVal["InputData"] = classAdJob.lookupAttribute("InputData")
        retVal["RescheduleCounter"] = reschedule_counter
        retVal["Status"] = JobStatus.RECEIVED
        retVal["MinorStatus"] = JobMinorStatus.RESCHEDULED

        return retVal

    async def get_job_status(self, job_id: int) -> LimitedJobStatusReturn:
        try:
            stmt = select(Jobs.Status, Jobs.MinorStatus, Jobs.ApplicationStatus).where(
                Jobs.JobID == job_id
            )
            return LimitedJobStatusReturn(
                **dict((await self.conn.execute(stmt)).one()._mapping)
            )
        except NoResultFound as e:
            raise JobNotFound(job_id) from e

    async def set_job_command(self, job_id: int, command: str, arguments: str = ""):
        """Store a command to be passed to the job together with the next heart beat"""
        try:
            stmt = insert(JobCommands).values(
                JobID=job_id,
                Command=command,
                Arguments=arguments,
                ReceptionTime=datetime.now(tz=timezone.utc),
            )
            await self.conn.execute(stmt)
        except IntegrityError as e:
            raise JobNotFound(job_id) from e

    async def delete_jobs(self, job_ids: list[int]):
        """
        Delete jobs from the database
        """
        stmt = delete(JobJDLs).where(JobJDLs.JobID.in_(job_ids))
        await self.conn.execute(stmt)

    async def set_properties(
        self, properties: dict[int, dict[str, Any]], update_timestamp: bool = False
    ) -> int:
        """Update the job parameters
        All the jobs must update the same properties

        :param properties: {job_id : {prop1: val1, prop2:val2}
        :param update_timestamp: if True, update the LastUpdate to now

        :return rowcount

        """

        # Check that all we always update the same set of properties
        required_parameters_set = set(
            [tuple(sorted(k.keys())) for k in properties.values()]
        )

        if len(required_parameters_set) != 1:
            raise NotImplementedError(
                "All the jobs should update the same set of properties"
            )

        required_parameters = list(required_parameters_set)[0]
        update_parameters = [{"job_id": k, **v} for k, v in properties.items()]

        columns = _get_columns(Jobs.__table__, required_parameters)
        values: dict[str, BindParameter[Any] | datetime] = {
            c.name: bindparam(c.name) for c in columns
        }
        if update_timestamp:
            values["LastUpdateTime"] = datetime.now(tz=timezone.utc)

        stmt = update(Jobs).where(Jobs.JobID == bindparam("job_id")).values(**values)
        rows = await self.conn.execute(stmt, update_parameters)

        return rows.rowcount


MAGIC_EPOC_NUMBER = 1270000000


class JobLoggingDB(BaseSQLDB):
    """Frontend for the JobLoggingDB. Provides the ability to store changes with timestamps"""

    metadata = JobLoggingDBBase.metadata

    async def insert_record(
        self,
        job_id: int,
        status: JobStatus,
        minor_status: str,
        application_status: str,
        date: datetime,
        source: str,
    ):
        """
        Add a new entry to the JobLoggingDB table. One, two or all the three status
        components (status, minorStatus, applicationStatus) can be specified.
        Optionally the time stamp of the status can
        be provided in a form of a string in a format '%Y-%m-%d %H:%M:%S' or
        as datetime.datetime object. If the time stamp is not provided the current
        UTC time is used.
        """

        # First, fetch the maximum SeqNum for the given job_id
        seqnum_stmt = select(func.coalesce(func.max(LoggingInfo.SeqNum) + 1, 1)).where(
            LoggingInfo.JobID == job_id
        )
        seqnum = await self.conn.scalar(seqnum_stmt)

        epoc = (
            time.mktime(date.timetuple())
            + date.microsecond / 1000000.0
            - MAGIC_EPOC_NUMBER
        )

        stmt = insert(LoggingInfo).values(
            JobID=int(job_id),
            SeqNum=seqnum,
            Status=status,
            MinorStatus=minor_status,
            ApplicationStatus=application_status[:255],
            StatusTime=date,
            StatusTimeOrder=epoc,
            Source=source[:32],
        )
        await self.conn.execute(stmt)

    async def get_records(self, job_id: int) -> list[JobStatusReturn]:
        """Returns a Status,MinorStatus,ApplicationStatus,StatusTime,Source tuple
        for each record found for job specified by its jobID in historical order
        """

        stmt = (
            select(
                LoggingInfo.Status,
                LoggingInfo.MinorStatus,
                LoggingInfo.ApplicationStatus,
                LoggingInfo.StatusTime,
                LoggingInfo.Source,
            )
            .where(LoggingInfo.JobID == int(job_id))
            .order_by(LoggingInfo.StatusTimeOrder, LoggingInfo.StatusTime)
        )
        rows = await self.conn.execute(stmt)

        values = []
        for (
            status,
            minor_status,
            application_status,
            status_time,
            status_source,
        ) in rows:
            values.append(
                [
                    status,
                    minor_status,
                    application_status,
                    status_time.replace(tzinfo=timezone.utc),
                    status_source,
                ]
            )

        # If no value has been set for the application status in the first place,
        # We put this status to unknown
        res = []
        if values:
            if values[0][2] == "idem":
                values[0][2] = "Unknown"

            # We replace "idem" values by the value previously stated
            for i in range(1, len(values)):
                for j in range(3):
                    if values[i][j] == "idem":
                        values[i][j] = values[i - 1][j]

            # And we replace arrays with tuples
            for (
                status,
                minor_status,
                application_status,
                status_time,
                status_source,
            ) in values:
                res.append(
                    JobStatusReturn(
                        Status=status,
                        MinorStatus=minor_status,
                        ApplicationStatus=application_status,
                        StatusTime=status_time,
                        Source=status_source,
                    )
                )

        return res

    async def delete_records(self, job_ids: list[int]):
        """Delete logging records for given jobs"""
        stmt = delete(LoggingInfo).where(LoggingInfo.JobID.in_(job_ids))
        await self.conn.execute(stmt)

    async def get_wms_time_stamps(self, job_id):
        """Get TimeStamps for job MajorState transitions
        return a {State:timestamp} dictionary
        """

        result = {}
        stmt = select(
            LoggingInfo.Status,
            LoggingInfo.StatusTimeOrder,
        ).where(LoggingInfo.JobID == job_id)
        rows = await self.conn.execute(stmt)
        if not rows.rowcount:
            raise JobNotFound(job_id) from None

        for event, etime in rows:
            result[event] = str(etime + MAGIC_EPOC_NUMBER)

        return result


class TaskQueueDB(BaseSQLDB):
    metadata = TaskQueueDBBase.metadata

    async def get_tq_infos_for_jobs(
        self, job_ids: list[int]
    ) -> set[tuple[int, str, str, str]]:
        """
        Get the task queue info for given jobs
        """
        stmt = (
            select(
                TaskQueues.TQId, TaskQueues.Owner, TaskQueues.OwnerGroup, TaskQueues.VO
            )
            .join(JobsQueue, TaskQueues.TQId == JobsQueue.TQId)
            .where(JobsQueue.JobId.in_(job_ids))
        )
        return set(
            (int(row[0]), str(row[1]), str(row[2]), str(row[3]))
            for row in (await self.conn.execute(stmt)).all()
        )

    async def get_owner_for_task_queue(self, tq_id: int) -> dict[str, str]:
        """
        Get the owner and owner group for a task queue
        """
        stmt = select(TaskQueues.Owner, TaskQueues.OwnerGroup, TaskQueues.VO).where(
            TaskQueues.TQId == tq_id
        )
        return dict((await self.conn.execute(stmt)).one()._mapping)

    async def remove_job(self, job_id: int):
        """
        Remove a job from the task queues
        """
        stmt = delete(JobsQueue).where(JobsQueue.JobId == job_id)
        await self.conn.execute(stmt)

    async def remove_jobs(self, job_ids: list[int]):
        """
        Remove jobs from the task queues
        """
        stmt = delete(JobsQueue).where(JobsQueue.JobId.in_(job_ids))
        await self.conn.execute(stmt)

    async def delete_task_queue_if_empty(
        self,
        tq_id: int,
        tq_owner: str,
        tq_group: str,
        job_share: int,
        group_properties: set[SecurityProperty],
        enable_shares_correction: bool,
        allow_background_tqs: bool,
    ):
        """
        Try to delete a task queue if it's empty
        """
        # Check if the task queue is empty
        stmt = (
            select(TaskQueues.TQId)
            .where(TaskQueues.Enabled >= 1)
            .where(TaskQueues.TQId == tq_id)
            .where(~TaskQueues.TQId.in_(select(JobsQueue.TQId)))
        )
        rows = await self.conn.execute(stmt)
        if not rows.rowcount:
            return

        # Deleting the task queue (the other tables will be deleted in cascade)
        stmt = delete(TaskQueues).where(TaskQueues.TQId == tq_id)
        await self.conn.execute(stmt)

        await self.recalculate_tq_shares_for_entity(
            tq_owner,
            tq_group,
            job_share,
            group_properties,
            enable_shares_correction,
            allow_background_tqs,
        )

    async def recalculate_tq_shares_for_entity(
        self,
        owner: str,
        group: str,
        job_share: int,
        group_properties: set[SecurityProperty],
        enable_shares_correction: bool,
        allow_background_tqs: bool,
    ):
        """
        Recalculate the shares for a user/userGroup combo
        """
        if JOB_SHARING in group_properties:
            # If group has JobSharing just set prio for that entry, user is irrelevant
            return await self.__set_priorities_for_entity(
                owner, group, job_share, group_properties, allow_background_tqs
            )

        stmt = (
            select(TaskQueues.Owner, func.count(TaskQueues.Owner))
            .where(TaskQueues.OwnerGroup == group)
            .group_by(TaskQueues.Owner)
        )
        rows = await self.conn.execute(stmt)
        # make the rows a list of tuples
        # Get owners in this group and the amount of times they appear
        # TODO: I guess the rows are already a list of tupes
        # maybe refactor
        data = [(r[0], r[1]) for r in rows if r]
        numOwners = len(data)
        # If there are no owners do now
        if numOwners == 0:
            return
        # Split the share amongst the number of owners
        entities_shares = {row[0]: job_share / numOwners for row in data}

        # TODO: implement the following
        # If corrector is enabled let it work it's magic
        # if enable_shares_correction:
        #     entities_shares = await self.__shares_corrector.correct_shares(
        #         entitiesShares, group=group
        #     )

        # Keep updating
        owners = dict(data)
        # IF the user is already known and has more than 1 tq, the rest of the users don't need to be modified
        # (The number of owners didn't change)
        if owner in owners and owners[owner] > 1:
            await self.__set_priorities_for_entity(
                owner,
                group,
                entities_shares[owner],
                group_properties,
                allow_background_tqs,
            )
            return
        # Oops the number of owners may have changed so we recalculate the prio for all owners in the group
        for owner in owners:
            await self.__set_priorities_for_entity(
                owner,
                group,
                entities_shares[owner],
                group_properties,
                allow_background_tqs,
            )

    async def __set_priorities_for_entity(
        self,
        owner: str,
        group: str,
        share,
        properties: set[SecurityProperty],
        allow_background_tqs: bool,
    ):
        """
        Set the priority for a user/userGroup combo given a splitted share
        """
        from DIRAC.WorkloadManagementSystem.DB.TaskQueueDB import calculate_priority

        stmt = (
            select(
                TaskQueues.TQId,
                func.sum(JobsQueue.RealPriority) / func.count(JobsQueue.RealPriority),
            )
            .join(JobsQueue, TaskQueues.TQId == JobsQueue.TQId)
            .where(TaskQueues.OwnerGroup == group)
            .group_by(TaskQueues.TQId)
        )
        if JOB_SHARING not in properties:
            stmt = stmt.where(TaskQueues.Owner == owner)
        rows = await self.conn.execute(stmt)
        tq_dict: dict[int, float] = {tq_id: priority for tq_id, priority in rows}

        if not tq_dict:
            return

        rows = await self.retrieve_task_queues(list(tq_dict))

        prio_dict = calculate_priority(tq_dict, rows, share, allow_background_tqs)

        # Execute updates
        for prio, tqs in prio_dict.items():
            update_stmt = (
                update(TaskQueues).where(TaskQueues.TQId.in_(tqs)).values(Priority=prio)
            )
            await self.conn.execute(update_stmt)

    async def retrieve_task_queues(self, tq_id_list=None):
        """
        Get all the task queues
        """
        if tq_id_list is not None and not tq_id_list:
            # Empty list => Fast-track no matches
            return {}

        stmt = (
            select(
                TaskQueues.TQId,
                TaskQueues.Priority,
                func.count(JobsQueue.TQId).label("Jobs"),
                TaskQueues.Owner,
                TaskQueues.OwnerGroup,
                TaskQueues.VO,
                TaskQueues.CPUTime,
            )
            .join(JobsQueue, TaskQueues.TQId == JobsQueue.TQId)
            .join(SitesQueue, TaskQueues.TQId == SitesQueue.TQId)
            .join(GridCEsQueue, TaskQueues.TQId == GridCEsQueue.TQId)
            .group_by(
                TaskQueues.TQId,
                TaskQueues.Priority,
                TaskQueues.Owner,
                TaskQueues.OwnerGroup,
                TaskQueues.VO,
                TaskQueues.CPUTime,
            )
        )
        if tq_id_list is not None:
            stmt = stmt.where(TaskQueues.TQId.in_(tq_id_list))

        tq_data: dict[int, dict[str, list[str]]] = dict(
            dict(row._mapping) for row in await self.conn.execute(stmt)
        )
        # TODO: the line above should be equivalent to the following commented code, check this is the case
        # for record in rows:
        #     tqId = record[0]
        #     tqData[tqId] = {
        #         "Priority": record[1],
        #         "Jobs": record[2],
        #         "Owner": record[3],
        #         "OwnerGroup": record[4],
        #         "VO": record[5],
        #         "CPUTime": record[6],
        #     }

        for tq_id in tq_data:
            # TODO: maybe factorize this handy tuple list
            for table, field in {
                (SitesQueue, "Sites"),
                (GridCEsQueue, "GridCEs"),
                (BannedSitesQueue, "BannedSites"),
                (PlatformsQueue, "Platforms"),
                (JobTypesQueue, "JobTypes"),
                (TagsQueue, "Tags"),
            }:
                stmt = select(table.Value).where(table.TQId == tq_id)
                tq_data[tq_id][field] = list(
                    row[0] for row in await self.conn.execute(stmt)
                )

        return tq_data



from __future__ import annotations

from diracx.db.os.utils import BaseOSDB


class JobParametersDB(BaseOSDB):
    fields = {
        "JobID": {"type": "long"},
        "timestamp": {"type": "date"},
        "CPUNormalizationFactor": {"type": "long"},
        "NormCPUTime(s)": {"type": "long"},
        "Memory(kB)": {"type": "long"},
        "TotalCPUTime(s)": {"type": "long"},
        "MemoryUsed(kb)": {"type": "long"},
        "HostName": {"type": "keyword"},
        "GridCE": {"type": "keyword"},
        "ModelName": {"type": "keyword"},
        "Status": {"type": "keyword"},
        "JobType": {"type": "keyword"},
    }
    index_prefix = "mysetup_elasticjobparameters_index_"

    def index_name(self, doc_id: int) -> str:
        # TODO: Remove setup and replace "123.0m" with "120m"?
        return f"{self.index_prefix}_{doc_id // 1e6:.1f}m"



from __future__ import annotations

import asyncio
from uuid import uuid4

import pytest

from diracx.core.exceptions import InvalidQueryError
from diracx.db.sql.dummy.db import DummyDB
from diracx.db.sql.utils import SQLDBUnavailable

# Each DB test class must defined a fixture looking like this one
# It allows to get an instance of an in memory DB,


@pytest.fixture
async def dummy_db(tmp_path) -> DummyDB:
    dummy_db = DummyDB("sqlite+aiosqlite:///:memory:")
    async with dummy_db.engine_context():
        async with dummy_db.engine.begin() as conn:
            await conn.run_sync(dummy_db.metadata.create_all)
        yield dummy_db


async def test_insert_and_summary(dummy_db: DummyDB):
    # Each context manager creates a transaction
    # So it is important to write test this way
    async with dummy_db as dummy_db:
        # First we check that the DB is empty
        result = await dummy_db.summary(["model"], [])
        assert not result

    # Now we add some data in the DB
    async with dummy_db as dummy_db:
        # Add a car owner
        owner_id = await dummy_db.insert_owner(name="Magnum")
        assert owner_id

        # Add cars, belonging to the same guy
        result = await asyncio.gather(
            *(dummy_db.insert_car(uuid4(), f"model_{i}", owner_id) for i in range(10))
        )
        assert result

    # Check that there are now 10 cars assigned to a single driver
    async with dummy_db as dummy_db:
        result = await dummy_db.summary(["ownerID"], [])

        assert result[0]["count"] == 10

    # Test the selection
    async with dummy_db as dummy_db:
        result = await dummy_db.summary(
            ["ownerID"], [{"parameter": "model", "operator": "eq", "value": "model_1"}]
        )

        assert result[0]["count"] == 1

    async with dummy_db as dummy_db:
        with pytest.raises(InvalidQueryError):
            result = await dummy_db.summary(
                ["ownerID"],
                [
                    {
                        "parameter": "model",
                        "operator": "BADSELECTION",
                        "value": "model_1",
                    }
                ],
            )


async def test_bad_connection():
    dummy_db = DummyDB("mysql+aiomysql://tata:yoyo@db.invalid:3306/name")
    async with dummy_db.engine_context():
        with pytest.raises(SQLDBUnavailable):
            async with dummy_db:
                dummy_db.ping()



import asyncio
from datetime import datetime, timezone
from unittest.mock import MagicMock

from fastapi import BackgroundTasks

from diracx.core.config.schema import Config
from diracx.core.exceptions import JobNotFound
from diracx.core.models import (
    JobStatus,
    JobStatusUpdate,
    ScalarSearchOperator,
    SetJobStatusReturn,
)
from diracx.db.sql.jobs.db import JobDB, JobLoggingDB, TaskQueueDB
from diracx.db.sql.sandbox_metadata.db import SandboxMetadataDB


async def set_job_status(
    job_id: int,
    status: dict[datetime, JobStatusUpdate],
    job_db: JobDB,
    job_logging_db: JobLoggingDB,
    force: bool = False,
) -> SetJobStatusReturn:
    """Set various status fields for job specified by its jobId.
    Set only the last status in the JobDB, updating all the status
    logging information in the JobLoggingDB. The status dict has datetime
    as a key and status information dictionary as values

    :raises: JobNotFound if the job is not found in one of the DBs
    """

    from DIRAC.Core.Utilities import TimeUtilities
    from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
    from DIRAC.WorkloadManagementSystem.Utilities.JobStatusUtility import (
        getNewStatus,
        getStartAndEndTime,
    )

    # transform JobStateUpdate objects into dicts
    statusDict = {}
    for key, value in status.items():
        statusDict[key] = {k: v for k, v in value.dict().items() if v is not None}

    res = await job_db.search(
        parameters=["Status", "StartExecTime", "EndExecTime"],
        search=[
            {
                "parameter": "JobID",
                "operator": ScalarSearchOperator.EQUAL,
                "value": str(job_id),
            }
        ],
        sorts=[],
    )
    if not res:
        raise JobNotFound(job_id) from None

    currentStatus = res[0]["Status"]
    startTime = res[0]["StartExecTime"]
    endTime = res[0]["EndExecTime"]

    # If the current status is Stalled and we get an update, it should probably be "Running"
    if currentStatus == JobStatus.STALLED:
        currentStatus = JobStatus.RUNNING

    # Get the latest time stamps of major status updates
    result = await job_logging_db.get_wms_time_stamps(job_id)

    #####################################################################################################

    # This is more precise than "LastTime". timeStamps is a sorted list of tuples...
    timeStamps = sorted((float(t), s) for s, t in result.items())
    lastTime = TimeUtilities.fromEpoch(timeStamps[-1][0]).replace(tzinfo=timezone.utc)

    # Get chronological order of new updates
    updateTimes = sorted(statusDict)

    newStartTime, newEndTime = getStartAndEndTime(
        startTime, endTime, updateTimes, timeStamps, statusDict
    )

    job_data = {}
    if updateTimes[-1] >= lastTime:
        new_status, new_minor, new_application = returnValueOrRaise(
            getNewStatus(
                job_id,
                updateTimes,
                lastTime,
                statusDict,
                currentStatus,
                force,
                MagicMock(),
            )
        )

        if new_status:
            job_data["Status"] = new_status
            job_data["LastUpdateTime"] = datetime.now(timezone.utc)
        if new_minor:
            job_data["MinorStatus"] = new_minor
        if new_application:
            job_data["ApplicationStatus"] = new_application

        # TODO: implement elasticJobParametersDB ?
        # if cls.elasticJobParametersDB:
        #     result = cls.elasticJobParametersDB.setJobParameter(int(jobID), "Status", status)
        #     if not result["OK"]:
        #         return result

    for updTime in updateTimes:
        if statusDict[updTime]["Source"].startswith("Job"):
            job_data["HeartBeatTime"] = updTime

    if not startTime and newStartTime:
        job_data["StartExecTime"] = newStartTime

    if not endTime and newEndTime:
        job_data["EndExecTime"] = newEndTime

    if job_data:
        await job_db.setJobAttributes(job_id, job_data)

    for updTime in updateTimes:
        sDict = statusDict[updTime]
        if not sDict.get("Status"):
            sDict["Status"] = "idem"
        if not sDict.get("MinorStatus"):
            sDict["MinorStatus"] = "idem"
        if not sDict.get("ApplicationStatus"):
            sDict["ApplicationStatus"] = "idem"
        if not sDict.get("Source"):
            sDict["Source"] = "Unknown"

        await job_logging_db.insert_record(
            job_id,
            sDict["Status"],
            sDict["MinorStatus"],
            sDict["ApplicationStatus"],
            updTime,
            sDict["Source"],
        )

    return SetJobStatusReturn(**job_data)


class ForgivingTaskGroup(asyncio.TaskGroup):
    # Hacky way, check https://stackoverflow.com/questions/75250788/how-to-prevent-python3-11-taskgroup-from-canceling-all-the-tasks
    # Basically e're using this because we want to wait for all tasks to finish, even if one of them raises an exception
    def _abort(self):
        return None


async def delete_jobs(
    job_ids: list[int],
    config: Config,
    job_db: JobDB,
    job_logging_db: JobLoggingDB,
    task_queue_db: TaskQueueDB,
    background_task: BackgroundTasks,
):
    """
    "Delete" jobs by removing them from the task queues, set kill as a job command setting the job status to DELETED.
    :raises: BaseExceptionGroup[JobNotFound] for every job that was not found
    """

    await _remove_jobs_from_task_queue(job_ids, config, task_queue_db, background_task)
    # TODO: implement StorageManagerClient
    # returnValueOrRaise(StorageManagerClient().killTasksBySourceTaskID(job_ids))

    async with ForgivingTaskGroup() as task_group:
        for job_id in job_ids:
            task_group.create_task(job_db.set_job_command(job_id, "Kill"))

            task_group.create_task(
                set_job_status(
                    job_id,
                    {
                        datetime.now(timezone.utc): JobStatusUpdate(
                            Status=JobStatus.DELETED,
                            MinorStatus="Checking accounting",
                            Source="job_manager",
                        )
                    },
                    job_db,
                    job_logging_db,
                    force=True,
                )
            )


async def kill_jobs(
    job_ids: list[int],
    config: Config,
    job_db: JobDB,
    job_logging_db: JobLoggingDB,
    task_queue_db: TaskQueueDB,
    background_task: BackgroundTasks,
):
    """
    Kill jobs by removing them from the task queues, set kill as a job command and setting the job status to KILLED.
    :raises: BaseExceptionGroup[JobNotFound] for every job that was not found
    """
    await _remove_jobs_from_task_queue(job_ids, config, task_queue_db, background_task)
    # TODO: implement StorageManagerClient
    # returnValueOrRaise(StorageManagerClient().killTasksBySourceTaskID(job_ids))

    async with ForgivingTaskGroup() as task_group:
        for job_id in job_ids:
            task_group.create_task(job_db.set_job_command(job_id, "Kill"))
            task_group.create_task(
                set_job_status(
                    job_id,
                    {
                        datetime.now(timezone.utc): JobStatusUpdate(
                            Status=JobStatus.KILLED,
                            MinorStatus="Marked for termination",
                            Source="job_manager",
                        )
                    },
                    job_db,
                    job_logging_db,
                    force=True,
                )
            )

    # TODO: Consider using the code below instead, probably more stable but less performant
    # errors = []
    # for job_id in job_ids:
    #     try:
    #         await job_db.set_job_command(job_id, "Kill")
    #         await set_job_status(
    #             job_id,
    #             {
    #                 datetime.now(timezone.utc): JobStatusUpdate(
    #                     Status=JobStatus.KILLED,
    #                     MinorStatus="Marked for termination",
    #                     Source="job_manager",
    #                 )
    #             },
    #             job_db,
    #             job_logging_db,
    #             force=True,
    #         )
    #     except JobNotFound as e:
    #         errors.append(e)

    # if errors:
    #     raise BaseExceptionGroup("Some job ids were not found", errors)


async def remove_jobs(
    job_ids: list[int],
    config: Config,
    job_db: JobDB,
    job_logging_db: JobLoggingDB,
    sandbox_metadata_db: SandboxMetadataDB,
    task_queue_db: TaskQueueDB,
    background_task: BackgroundTasks,
):
    """
    Fully remove a job from the WMS databases.
    :raises: nothing
    """

    # Remove the staging task from the StorageManager
    # TODO: this was not done in the JobManagerHandler, but it was done in the kill method
    # I think it should be done here too
    # TODO: implement StorageManagerClient
    # returnValueOrRaise(StorageManagerClient().killTasksBySourceTaskID([job_id]))

    # TODO: this was also  not done in the JobManagerHandler, but it was done in the JobCleaningAgent
    # I think it should be done here as well
    await sandbox_metadata_db.unassign_sandbox_from_jobs(job_ids)

    # Remove the job from TaskQueueDB
    await _remove_jobs_from_task_queue(job_ids, config, task_queue_db, background_task)

    # Remove the job from JobLoggingDB
    await job_logging_db.delete_records(job_ids)

    # Remove the job from JobDB
    await job_db.delete_jobs(job_ids)


async def _remove_jobs_from_task_queue(
    job_ids: list[int],
    config: Config,
    task_queue_db: TaskQueueDB,
    background_task: BackgroundTasks,
):
    """
    Remove the job from TaskQueueDB
    """
    tq_infos = await task_queue_db.get_tq_infos_for_jobs(job_ids)
    await task_queue_db.remove_jobs(job_ids)
    for tq_id, owner, owner_group, vo in tq_infos:
        # TODO: move to Celery
        background_task.add_task(
            task_queue_db.delete_task_queue_if_empty,
            tq_id,
            owner,
            owner_group,
            config.Registry[vo].Groups[owner_group].JobShare,
            config.Registry[vo].Groups[owner_group].Properties,
            config.Operations[vo].Services.JobScheduling.EnableSharesCorrection,
            config.Registry[vo].Groups[owner_group].AllowBackgroundTQs,
        )



from __future__ import annotations

import asyncio
import secrets
from datetime import datetime

import pytest
import sqlalchemy

from diracx.core.models import SandboxInfo, UserInfo
from diracx.db.sql.sandbox_metadata.db import SandboxMetadataDB
from diracx.db.sql.sandbox_metadata.schema import sb_SandBoxes


@pytest.fixture
async def sandbox_metadata_db(tmp_path):
    sandbox_metadata_db = SandboxMetadataDB("sqlite+aiosqlite:///:memory:")
    async with sandbox_metadata_db.engine_context():
        async with sandbox_metadata_db.engine.begin() as conn:
            await conn.run_sync(sandbox_metadata_db.metadata.create_all)
        yield sandbox_metadata_db


def test_get_pfn(sandbox_metadata_db: SandboxMetadataDB):
    user_info = UserInfo(
        sub="vo:sub", preferred_username="user1", dirac_group="group1", vo="vo"
    )
    sandbox_info = SandboxInfo(
        checksum="checksum",
        checksum_algorithm="sha256",
        format="tar.bz2",
        size=100,
    )
    pfn = sandbox_metadata_db.get_pfn("bucket1", user_info, sandbox_info)
    assert pfn == "/S3/bucket1/vo/group1/user1/sha256:checksum.tar.bz2"


async def test_insert_sandbox(sandbox_metadata_db: SandboxMetadataDB):
    user_info = UserInfo(
        sub="vo:sub", preferred_username="user1", dirac_group="group1", vo="vo"
    )
    pfn1 = secrets.token_hex()

    # Make sure the sandbox doesn't already exist
    db_contents = await _dump_db(sandbox_metadata_db)
    assert pfn1 not in db_contents
    async with sandbox_metadata_db:
        with pytest.raises(sqlalchemy.exc.NoResultFound):
            await sandbox_metadata_db.sandbox_is_assigned("SandboxSE", pfn1)

    # Insert the sandbox
    async with sandbox_metadata_db:
        await sandbox_metadata_db.insert_sandbox("SandboxSE", user_info, pfn1, 100)
    db_contents = await _dump_db(sandbox_metadata_db)
    owner_id1, last_access_time1 = db_contents[pfn1]

    # Inserting again should update the last access time
    await asyncio.sleep(1)  # The timestamp only has second precision
    async with sandbox_metadata_db:
        await sandbox_metadata_db.insert_sandbox("SandboxSE", user_info, pfn1, 100)
    db_contents = await _dump_db(sandbox_metadata_db)
    owner_id2, last_access_time2 = db_contents[pfn1]
    assert owner_id1 == owner_id2
    assert last_access_time2 > last_access_time1

    # The sandbox still hasn't been assigned
    async with sandbox_metadata_db:
        assert not await sandbox_metadata_db.sandbox_is_assigned("SandboxSE", pfn1)

    # Inserting again should update the last access time
    await asyncio.sleep(1)  # The timestamp only has second precision
    last_access_time3 = (await _dump_db(sandbox_metadata_db))[pfn1][1]
    assert last_access_time2 == last_access_time3
    async with sandbox_metadata_db:
        await sandbox_metadata_db.update_sandbox_last_access_time("SandboxSE", pfn1)
    last_access_time4 = (await _dump_db(sandbox_metadata_db))[pfn1][1]
    assert last_access_time2 < last_access_time4


async def _dump_db(
    sandbox_metadata_db: SandboxMetadataDB,
) -> dict[str, tuple[int, datetime]]:
    """Dump the contents of the sandbox metadata database

    Returns a dict[pfn: str, (owner_id: int, last_access_time: datetime)]
    """
    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(
            sb_SandBoxes.SEPFN, sb_SandBoxes.OwnerId, sb_SandBoxes.LastAccessTime
        )
        res = await sandbox_metadata_db.conn.execute(stmt)
        return {row.SEPFN: (row.OwnerId, row.LastAccessTime) for row in res}
